/*
 * Defines for the PCF8574.
 *
 * There are two PCF8574As on the board.  One is at address 0x70 
 * (W: 0x70, R: 0x71) and performs various control functions to do with 
 * the RGB section:
 *
 * P7: saa7196 enable (0=en, 1=dis)
 * P6: bt254 enable (0=en, 1=dis)
 * P5: ntsc (1=ntsc, 0=pal/secam)
 * P4: BT254 WR*
 * P3: ?? (probably BT254 RD*)
 * P2: BT254 A2
 * P1: BT254 A1
 * P0: BT254 A0
 *
 * The other is the data for (or from) the Bt254 at address 0x72 (W: 0x72, R: 0x73)
 *
 */

#ifndef _BT_254_H
#define _BT_254_H

#define NUM_PCF8574_I2C_REGS	2
#define	PCF8574_CTRL_I2C_ADDR_W	0x70
#define	PCF8574_CTRL_I2C_ADDR_R	0x71
#define PCF8574_DATA_I2C_ADDR_W	0x72
#define PCF8574_DATA_I2C_ADDR_R	0x73
#define	PCF8574_CTRL_WRITE(mtr, data) \
 saa7116_i2c(mtr,  PCF8574_CTRL_I2C_ADDR_W, data, data), \
	mtr->pcf_i2c[0] = data
#define	PCF8574_DATA_WRITE(mtr, data) \
 saa7116_i2c(mtr,  PCF8574_DATA_I2C_ADDR_W, data, data), \
	mtr->pcf_i2c[1] = data
#define PCF8574_CTRL_REG(mtr) mtr->pcf_i2c[0]
#define PCF8574_DATA_REG(mtr) mtr->pcf_i2c[1]


/*
 * Defines for the BT254.
 */
#define	NUM_BT254_REGS	7

#define BT254_COMMAND	0
#define	BT254_IOUT1	1
#define	BT254_IOUT2	2
#define	BT254_IOUT3	3
#define BT254_IOUT4	4
#define	BT254_IOUT5	5
#define	BT254_IOUT6	6

/* Internal BT254 Registers */
static u_char bt254_default[NUM_BT254_REGS] =
{
  0x00,				/* 24 bpp */
  0xa0,
  0xa0,
  0xa0,
  0x50,
  0x50,
  0x50,
};



#endif
