/*      Video for Linux Two
 *      Matrox Meteor Capture Driver
 *
 *      This software is in the public domain.
 *      Written by Olivier Carmona (c) 1999-2000
 *      This driver is a complete rewrite of the Linux Meteor driver
 *      written by Jim Bray (c) 1996, already based on Mark Tinguely 
 *      and Jim Lowe FreeBSD Meteor driver.
 *      Credits to Bill Dirks for V4L2 Sample Code & for many hints.
 *      Credits to Hacene Machane from Meteor Matrox Imaging for SAA7116
 *                 datasheet.
 *      Credits to Petter Reinholdtsen for many patches.
 *      Credits to Mark Charter for YUV odd only bug.
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *      Comments and bug reports should be sent to Olivier Carmona
 *      <olivier.carmona@ieee.org>.
 *
 * ---------------------------------------------------------------
 *  
 *      Feb 2004 marco corvi <marco_corvi@geocities.com>
 *      porting to kernel 2.6 (tested with 2.6.3):
 *      
 *      Done:
 *      [1] use V4L2 in the mainline of the kernel
 *      [2] update to the new v4l2 struct's
 *      [3] use the 2.6 wait mechanism (old code preprocessed out)
 *      [4] changed interrupt_hw() return type
 *      [5] dropped support for Linux version < 2.4
 *      
 *      ToDo:
 *      [1] replace the custom queue with kernel generic list
 *          there is something in the code, it is working, but i am 
 *          not happy with it
 *      [2] fix the time management to use one or two units only
 *      [3] fix the possible values of "field"
 *      [4] defer capture_interrupt with a workqueue
 *          this is under way
 *      [5] use driver class model for ioctl's
 *
 *      Maybe:
 *      [1] why to keep the masterclock ?
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <linux/sched.h>

#include <linux/version.h>
#include <linux/moduleparam.h>

#include <linux/videodev.h>
#include <linux/videodev2.h>

#include <linux/version.h>
#include <asm/uaccess.h>
#include <linux/pci.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/interrupt.h>
#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif

/* --------------------------------------------------------- */
// use the new wait mechanism
// #define METEOR_NEW_WAIT 1

// use the 2.6 workqueue mechanism
// #define METEOR_USE_WORKQUEUE 1

// use the 2.6 struct timespec
#define METEOR_USE_TIMESPEC 1

// use the kernel list_head struct
// working, but not what i wanted
// #define METEOR_USE_LIST 1

/* --------------------------------------------------------- */

#include "saa71x6.h"
#include "bt254.h"
#include "meteor_v4l2.h"

MODULE_AUTHOR("oliver carmona");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Meteor Frame Grabber");

#ifndef PKMOD
#define PKMOD "mtr: "
#endif

// #define DEBUG

#ifdef DEBUG
#define debug_msg(fmt,arg...) printk(KERN_DEBUG PKMOD fmt,##arg)
#define debug_ioctl(fmt,arg...) printk(KERN_DEBUG PKMOD fmt,##arg)
#else
#define debug_msg(fmt,arg...)
#define debug_ioctl(fmt,arg...)
#endif

#define err_msg(fmt,arg...) printk(KERN_ERR PKMOD fmt,##arg)
#define info_msg(fmt,arg...) printk(KERN_INFO PKMOD fmt,##arg)


#ifdef METEOR_USE_WORKQUEUE
static void capture_interrupt( void * );
struct workqueue_struct * meteor_wq_thread;

DECLARE_WORK( meteor_wq, capture_interrupt, NULL );
#endif

/* ==============================================================
 *
 *    V4L2 extra stuff
 *
 * This is generic V4L2 stuff from the old code that is not in the kernel 
 *
 * -------------------------------------------------------------- */

static rwlock_t rw_lock_unlocked = RW_LOCK_UNLOCKED;
static struct v4l2_clock * masterclock;

#ifdef METEOR_USE_LIST

static inline int 
meteor_list_is_last( struct list_head * head) 
{
	unsigned long flags;
        read_lock_irqsave(&rw_lock_unlocked, flags);
	if ( ( head->next == head ) && ( head->prev == head ) ) {
        	read_unlock_irqrestore(&rw_lock_unlocked, flags);
		return -1;
	}
	if ( head->next == head->prev ) {
        	read_unlock_irqrestore(&rw_lock_unlocked, flags);
		return 1;
	}
        read_unlock_irqrestore(&rw_lock_unlocked, flags);
	return 0;
}

static inline struct stream_buffer * 
meteor_list_peek_head( struct list_head * head)
{
	unsigned long flags;
        read_lock_irqsave(&rw_lock_unlocked, flags);
  	if ( head->next == head ) {
        	read_unlock_irqrestore(&rw_lock_unlocked, flags);
    		return NULL;
	}
  	struct stream_buffer * buf =
	       	container_of( head->next, struct stream_buffer, qnode );
	read_unlock_irqrestore(&rw_lock_unlocked, flags);
	return buf;
}

static inline void
meteor_list_try_del( struct list_head * entry, struct list_head * head)
{
	unsigned long flags;
	struct list_head * tmp;

        read_lock_irqsave(&rw_lock_unlocked, flags);
	list_for_each( tmp, head )
		if ( tmp == entry ) {
			list_del( tmp );
			read_unlock_irqrestore(&rw_lock_unlocked, flags);
			return;
		}
	read_unlock_irqrestore(&rw_lock_unlocked, flags);
}

static inline void
meteor_list_add_tail( struct list_head * entry, struct list_head * head )
{
	unsigned long flags;
	read_lock_irqsave(&rw_lock_unlocked, flags);
	list_add_tail( entry, head );
	read_unlock_irqrestore(&rw_lock_unlocked, flags);
}

static inline void
meteor_list_unqueue( struct list_head * head )
{
	struct list_head * tmp;
	unsigned long flags;
	struct stream_buffer * buf;

	read_lock_irqsave(&rw_lock_unlocked, flags);
        list_for_each( tmp, head ) {
	  list_del( tmp );
	  buf = container_of( tmp, struct stream_buffer, qnode);
	  buf->vidbuf.flags &= ~V4L2_BUF_FLAG_QUEUED;
	}
	read_unlock_irqrestore(&rw_lock_unlocked, flags);
}

static inline void
meteor_list_move_head_to_tail ( struct list_head * from, struct list_head * to)
{
	unsigned long flags;
	struct list_head * tmp;
	       
	read_lock_irqsave(&rw_lock_unlocked, flags);
	tmp = from->next;
	if ( tmp == from ) {
		read_unlock_irqrestore(&rw_lock_unlocked, flags);
		return;
	}
	list_del ( tmp );
	list_add_tail( tmp, to );
	read_unlock_irqrestore(&rw_lock_unlocked, flags);
}

static inline struct stream_buffer *
meteor_list_pop_head( struct list_head * head )
{
	unsigned long flags;
	struct stream_buffer * buf;
	struct list_head * tmp;

	read_lock_irqsave(&rw_lock_unlocked, flags);
	tmp = head->next;
	if ( tmp == head ) {
		read_unlock_irqrestore(&rw_lock_unlocked, flags);
		return NULL;
	}
	buf = container_of( tmp, struct stream_buffer, qnode);
        list_del( tmp );
	read_unlock_irqrestore(&rw_lock_unlocked, flags);
	return buf;
}

#else


int
v4l2_q_last(struct v4l2_queue *q)
{
/*  This function by Olivier Carmona  */

    unsigned long flags;
    read_lock_irqsave(&(q->qlock), flags);
    if (q == NULL)
    {
        read_unlock_irqrestore(&(q->qlock), flags);
        return -1;
    }
    if (q->forw == NULL || q->back == NULL ||
        q->forw == (struct v4l2_q_node *)q ||
        q->back == (struct v4l2_q_node *)q)
    {
        read_unlock_irqrestore(&(q->qlock), flags);
        return -1;
    }
    if (q->forw == q->back)
    {
        read_unlock_irqrestore(&(q->qlock), flags);
        return 1;
    }
    read_unlock_irqrestore(&(q->qlock), flags);
    return 0;
}



void 
v4l2_q_init( struct v4l2_queue * q ) 
{
	if ( NULL == q )
		return;
	q->qlock = rw_lock_unlocked;
	q->forw  = (struct v4l2_q_node *) q;
	q->back  = (struct v4l2_q_node *) q;
}

void
v4l2_q_add_tail( struct v4l2_queue * q, struct v4l2_q_node * node )
{
	unsigned long flags;
	if (q == NULL || node == NULL)
		return;
	if (q->forw == NULL || q->back == NULL)
		v4l2_q_init(q);
	write_lock_irqsave(&(q->qlock), flags);
	node->forw = (struct v4l2_q_node *)q;
	node->back = q->back;
	q->back->forw = node;
	q->back = node;
	write_unlock_irqrestore(&(q->qlock), flags);
}

void *
v4l2_q_del_head( struct v4l2_queue * q )
{
	unsigned long flags;
    struct v4l2_q_node *node;
    if (q == NULL)
        return NULL;
    write_lock_irqsave(&(q->qlock), flags);
    if (q->forw == NULL || q->back == NULL ||
        q->forw == (struct v4l2_q_node *)q ||
        q->back == (struct v4l2_q_node *)q)
    {
        write_unlock_irqrestore(&(q->qlock), flags);
        return NULL;
    }
    node = q->forw;
    node->forw->back = (struct v4l2_q_node *)q;
    q->forw = node->forw;
    node->forw = NULL;
    node->back = NULL;
    write_unlock_irqrestore(&(q->qlock), flags);
    return node;
}

/** get the next item on the queue
 * @param q  current item
 * @return   the item after "q", or NULL if there is none
 */ 
void *
v4l2_q_peek_head(struct v4l2_queue *q)
{
	unsigned long flags;
	struct v4l2_q_node *node;
	read_lock_irqsave(&(q->qlock), flags);
	if (q == NULL || q->forw == NULL || q->forw == (struct v4l2_q_node *)q)
	{
		read_unlock_irqrestore(&(q->qlock), flags);
		return NULL;
	}
	node = q->forw;
	read_unlock_irqrestore(&(q->qlock), flags);
	return node;
}


void *
v4l2_q_yank_node(struct v4l2_queue *q, struct v4l2_q_node *node)
{
    unsigned long flags;
    struct v4l2_q_node *t;
    if (v4l2_q_peek_head(q) == NULL || node == NULL)
        return NULL;
    write_lock_irqsave(&(q->qlock), flags);
    for (t = q->forw; t != (struct v4l2_q_node *)q; t = t->forw)
        if (t == node)
        {
            node->back->forw = node->forw;
            node->forw->back = node->back;
            node->forw = NULL;
            node->back = NULL;
            write_unlock_irqrestore(&(q->qlock), flags);
            return node;
        }
    write_unlock_irqrestore(&(q->qlock), flags);
    return NULL;
}

#endif /* METEOR_USE_LIST */

/*
 *  Math functions
 */

u32
v4l2_math_div6432(u64 a, u32 d, u32 *r)
{
    u32 q, m;
#ifdef __i386__
/*  Danger: This function will fault if the quotient exceeds (1<<32) - 1  */
    __asm__ __volatile__ (
        "   movl %2,%%eax\n"
        "   movl %3,%%edx\n"
        "   divl %4\n"
        "   movl %%eax,%0\n"
        "   movl %%edx,%1\n"
        : "=g" (q), "=g" (m)
        : "g" ((u32)a), "g" ((u32)(a >> 32)), "g" (d)
        : "eax", "edx"
        );
#else
    q = a / d;
    m = a % d;
#endif
    if (r) *r = m;
    return q;
}

unsigned long
v4l2_timestamp_divide(stamp_t t, unsigned long p_100ns )
{
        /*  Note: 't' is in 1ns units, 'p_100ns' is in 100ns units, */
        /*  and the quotient is rounded  */
        u64     p;

        p = (u64)p_100ns * 100;  /* 1ns units */
        t >>= 6;      /*  /64 to allow p_100ns longer than 4 secs. */
        p >>= 6;      /*  to keep quotient the same  */
        return v4l2_math_div6432((u64)t + (p >> 1), (u32)p, NULL);
}

/** Force the timestamp to be an integer multiple of p_100ns  
 *
 * @param   t  timestamp (units ns)
 * @param   p_100ns timeperframe (units 100ns)
 * @return  the number of frames, N = t/(100 * p_100ns)
 *
 *   N.B. side-effect: set timestamp t = N * p_100ns * 100 
 *   [the factor 100 accounts for units: 100ns -> ns]
 */
unsigned long
v4l2_timestamp_correct(stamp_t *t, unsigned long p_100ns)
{
        /*  Note: 't' is in 1ns units, 'p_100ns' is in 100ns units */
        unsigned long   n;

        n = v4l2_timestamp_divide((u64)*t, p_100ns);
        *t = (u64)p_100ns * n * 100;
        return n;
}

/** Get the current time
 * @param curr output timestamp (in nanoseconds)
 */
void
v4l2_masterclock_gettime(stamp_t *curr)
{
    if (masterclock)
        masterclock->gettime(curr);
    else
    {
#if defined(CONFIG_UST) || defined(CONFIG_UST_MODULE)
        ust_gettime(curr);
#else
        // (mc) using this in my config
        struct timeval  t;
        stamp_t stamp;
        do_gettimeofday(&t);
        stamp = (stamp_t)t.tv_sec * 1000000 + t.tv_usec;
        stamp *= 1000;
        *curr = stamp;
#endif
    }
}

/* Verify the validity of the parameters of a v4l2_standard structure and
 *   create the name and id from the other fields.  It does not relieve a 
 *   driver from examining if it can fulfill the request.  Returns an 
 *   errno < 0 if inconsistent, 0 if an unknown but maybe usable format, 
 *   or the V4L2_STD_XXX_X value if a known standard. 
 */
int
v4l2_video_std_confirm(struct v4l2_standard *vs)
{
    unsigned int    rate  = 0;
    unsigned int    lines = vs->framelines;

    strcpy(vs->name, "Unknown");
    if ( vs->reserved[0] || vs->reserved[1] || vs->reserved[2] || vs->reserved[3] )
        return -EINVAL;

    if (vs->frameperiod.numerator > 0 &&
        vs->frameperiod.denominator > 0)
        rate = v4l2_video_std_fps(vs);

    if (vs->framelines >= 624 && vs->framelines <= 626)
        lines = 625;
    else if (vs->framelines >= 524 && vs->framelines <= 526)
        lines = 525;

    if (rate == 0 || lines == 0 || rate > 200)
        return -EINVAL;

    if (vs->id & V4L2_STD_PAL) {
        strcpy(vs->name, "PAL");
    } else if (vs->id & V4L2_STD_NTSC) {
        strcpy(vs->name, "NTSC");
    } else if (vs->id & V4L2_STD_SECAM) {
        strcpy(vs->name, "SECAM");
    } else {
        return -EINVAL;
    }

    sprintf(vs->name + strlen(vs->name), " %d/%d", vs->framelines, rate);

    return 0;
}

static int
v4l2_fract_isless( struct v4l2_fract f1, struct v4l2_fract f2 )
{
	__u64 prod1 = f1.numerator * f2.denominator;
	__u64 prod2 = f2.numerator * f1.denominator;
	return ( prod1 < prod2 ) ? 1 : 0;
}

static unsigned long
v4l2_fract_to_100ns( struct v4l2_fract f )
{
	__u64 tmp = f.numerator * 10000000;
	return v4l2_math_div6432( tmp, f.denominator, NULL);
}


/* --------------------------------------------------------------
 *
 * End V4L2 extra stuff
 *
 * ============================================================== */

/*  Video controls                                                    */
/*  fields: id, type, name, min, max, step, default, flags, reserved  */

static struct v4l2_queryctrl capture_control[] =
{
  {V4L2_CID_BRIGHTNESS, V4L2_CTRL_TYPE_INTEGER, "Brightness", 0, 255, 1, 0x80},
  {V4L2_CID_CONTRAST, V4L2_CTRL_TYPE_INTEGER, "Contrast", 0, 127, 1, 0x40},
  {V4L2_CID_SATURATION, V4L2_CTRL_TYPE_INTEGER, "Saturation", 0, 127, 1, 0x40},
  {V4L2_CID_HUE, V4L2_CTRL_TYPE_INTEGER, "Hue", 0, 255, 1, 0x80},
  {V4L2_CID_GAIN, V4L2_CTRL_TYPE_INTEGER, "Gain", 0, 255, 1, 0x2c},
                                  /* 0x2c: CCIR Level for NTSC */
                                  /* 0x59: CCIR Level for PAL */
#if 1				/* extra stuff used for testing vidpanel */
  {V4L2_CID_EXPOSURE, V4L2_CTRL_TYPE_MENU, "Exposure", 0, 4, 0, 0},
  {V4L2_CID_AUTOGAIN, V4L2_CTRL_TYPE_BOOLEAN, "Auto Gain", 0, 1, 0, 1},
  {V4L2_CID_DO_WHITE_BALANCE, V4L2_CTRL_TYPE_BUTTON, "White Balance", 0, 0, 0, 0},
#endif
};

#define VCTRL_BRIGHTNESS	0
#define VCTRL_CONTRAST		1
#define VCTRL_SATURATION	2
#define VCTRL_HUE		3
#define VCTRL_GAIN		4
#define VCTRL_GAMMA		5

static int
find_vctrl (int id)
{
  int i;
  if (id == V4L2_CID_PRIVATE_BASE ||
      id < V4L2_CID_BASE ||
      id > V4L2_CID_LASTP1)
    return -EDOM;
  for (i = MAXCONTROLS - 1; i >= 0; i--)
    if (capture_control[i].id == id)
      break;
  if (i < 0)
    i = -EINVAL;
  return i;
}
static int
vctrl_querymenu (struct v4l2_querymenu *qm)
{
  static char *expo_menu[] =
  {
    "1/60",
    "1/100",
    "1/250",
    "1/1000",
    "1/5000",
  };
  if (qm->id == V4L2_CID_EXPOSURE)
    {
      if (qm->index < 0 ||
	  qm->index >= sizeof (expo_menu) / sizeof (char *))
	  return -EINVAL;
      {
	memcpy (qm->name, expo_menu[qm->index],
		sizeof (qm->name));
      }
      return 0;
    }
  return -EINVAL;
}


/*      Values for type field   */
#define DEVICE_TYPE_0		0

/*      Extreme video dimensions        */
#define MIN_WIDTH		16
#define MIN_HEIGHT		10
#define MAX_WIDTH		768
#define MAX_HEIGHT		576
#define MAX_FRAME_AGE		200	/* ms */

/*
 * The Capture device structure array. This is the only global
 *   variable in the module besides those used by the device probing
 *   and enumeration routines (command line overrides)
 */
#define MAXMETEORS		2
static struct meteor meteors[MAXMETEORS];
// static int unit_video[MAXMETEORS] = {0, 1};
static int input = METEOR_INPUT_DEV0;
static int standard = V4L2_STD_PAL;

// #ifdef MODULE_PARM
// MODULE_PARM (unit_video, "1-" __MODULE_STRING (MAXMETEORS) "i");
// MODULE_PARM(input, "i");
// MODULE_PARM(standard, "i");
// #endif
// module_param( unit_video, int, 0 );
module_param( input, int, 0 );
module_param( standard, int, 0 );


struct meteor *
meteor_from_file( struct inode * inode, struct file * file, struct device_open ** optr )
{
  struct meteor * mtr;
  *optr = (struct device_open *) file->private_data;
  if ( NULL == (*optr) ) 
  {
#ifdef DEBUG
	int minor = (inode != NULL) ? iminor( inode ) : -1;
	debug_msg("Invalid device_open. Minor %d\n", minor);
#endif
        return NULL;
  }
  mtr = (*optr)->dev;
  if ( NULL == mtr ) 
  {
#ifdef DEBUG
	int minor = (inode != NULL) ? iminor( inode ) : -1;
	debug_msg("Invalid meteor. Minor %d\n", minor);
#endif
        return NULL;
  }
  return mtr;
}


/*  These macros can be used to make device I/O operations atomic  
static spinlock_t device_lock = SPIN_LOCK_UNLOCKED; 
#define BEGIN_CRITICAL_SECTION       \
         do{__u32 flags;spin_lock_irqsave(&wavi_lock,flags);
#define END_CRITICAL_SECTION \
      spin_unlock_irqrestore(&wavi_lock,flags);}while(0)
*/

static int capture_new_format (struct meteor *mtr);

/*
 * Delay for WaitJiffies via timeout and schedule().
 * We need to be technically interruptable for this to work,
 * but we really don't want to be, so we block everything.
 * Don't call from interrupt.
 */

static void
wait_jiffies (__u32 waitJiffies)
{
  sigset_t saveBlocked = current->blocked;
  saveBlocked = current->blocked;
  sigemptyset (&current->blocked);
  current->state = TASK_INTERRUPTIBLE;
  schedule_timeout (waitJiffies);
  current->blocked = saveBlocked;
}

/* ==============================================================
 *
 *    S A A 7 1 1 6   P C I   B R I D G E   P R O C E D U R E S
 *
 * ============================================================== */

/*  I2C communication main routine */
static int
saa7116_i2c (struct meteor *mtr, __u8 dev, __u8 addr, __u8 data)
{
  __u32 start_jiffies = jiffies;
  register struct saa7116 *s7116 = mtr->s7116;

  /* execute i2c cycle */
  s7116->i2c_cmd = (struct i2c_cmd){
    .w_data   = data, 
    .reg_addr = addr, 
    .dev_addr = dev, 
    .xfer     = 1
  };


  /* and wait for completion */
  for (;;)
    {
      wait_jiffies (1);
      if (!s7116->i2c_cmd.xfer)
	break;
      if (jiffies - start_jiffies > WAIT_JIFFS)
	{
	  debug_msg ("meteor%d: i2c timeout error.\n", mtr->index);
	  return 1;
	}
    }

  if (s7116->i2c_stat.d_abort)
    {
      s7116->i2c_stat.d_abort = 1;	/* Reset bit to zero */
      debug_msg ("meteor%d: i2c abort error.\n", mtr->index);
      return 1;
    }

  return 0;
}


static void
saa7116_fps (struct meteor* mtr)
{
  register struct saa7116 *s7116 = mtr->s7116;

  s7116->field_mask_e =
    s7116->field_mask_o = 0x00000001;/* This combination of masks and */
  s7116->mask_lengths = 0;	/* mask length means capture all fields */
}

static void
saa7116_reset (struct meteor *mtr)
{
  register struct saa7116 *s7116 = mtr->s7116;

  wait_jiffies (WAIT_JIFFS);	/* Let things settle down */
  s7116->capt_ctl = 0x00000f30;	/* Reset Video Decoder */
  memset (s7116, 0, sizeof (struct saa7116));	/* Clear and reset everything */
  wait_jiffies (WAIT_JIFFS);	/* Let the dust settle */
}

static void
saa7116_init (struct meteor *mtr)
{
  struct saa7116 *s7116;
  __u32 s7116_page, s7116_off, s7116_addr;
  
  /* Map saa7196 regs into our address space */
  if (mtr->s7116 != NULL)
      iounmap ((void *) mtr->s7116);	/* Unwind ioremap() */
      
  pci_read_config_dword (mtr->pci, PCI_BASE_ADDRESS_0, &s7116_addr);

  s7116_page = s7116_addr & PAGE_MASK;
  s7116_off = s7116_addr & ~PAGE_MASK;

  s7116_addr = s7116_off | (__u32) ioremap ((__u32) s7116_page, sizeof (struct saa7116));

  if (!s7116_addr)
    {
      err_msg ("meteor%d: Mapping of SAA7196 failed\n", mtr->index);
      return;
    }

  mtr->s7116 = (struct saa7116 *) s7116_addr;
  s7116 = mtr->s7116;
  
  saa7116_reset (mtr);
  
  s7116->capt_ctl = 0x00008fc0;	/* Enable range check, clear errors, bring out 
                                   of reset 7196 & enable FIFO  */
  wait_jiffies (WAIT_JIFFS);
  s7116->i2c_clocks = 0x461e1e0f;
  s7116->rt_mode_e = s7116->rt_mode_o = (struct rt_mode){
    .mode  = 0x01, 
    .route = 0xeeeeee
  }; /* RGB 16 */
  s7116->fifo_trigger = (struct ff_trig){
    .packed = 0x20, 
    .planar = 0x20
  };
  s7116->field_toggle = 0x0000107; /* Disable field toggle */
  s7116->ints_ctl     = 0x07;	   /* Clear and disable all interrupts */
  s7116->field_mask_e =
    s7116->field_mask_o = 0x00000001;/* This combination of masks and */
  s7116->mask_lengths = 0;	   /* mask length means capture all fields */
  s7116->fifo_limits = 0x0005007c;
  wait_jiffies (WAIT_JIFFS);	/* Let this sink in */
}

static inline int
saa7116_check (struct meteor *mtr)
{
  __u16 status;
  __u32 cap_err;
  register struct saa7116 *s7116 = mtr->s7116;

  /* Check PCI Status */
  pci_read_config_word (mtr->pci, PCI_STATUS, &status);
  pci_write_config_word (mtr->pci, PCI_STATUS, status);		// Clear errors
  if (status & PCI_STATUS_REC_TARGET_ABORT)
   {
    debug_msg ("meteor%d: pci target abort\n", mtr->index);
    return 0;
   }
  if (status & PCI_STATUS_DETECTED_PARITY)
   {
    debug_msg ("meteor%d: pci parity error\n", mtr->index);
    return 0;
   }
  if (status & PCI_STATUS_REC_MASTER_ABORT)
   {
    debug_msg ("meteor%d: pci master abort\n", mtr->index);
    
    /* Check for FIFO and DMA address errors.*/
    cap_err = s7116->capt_ctl & 0x0f00;
  
    if (cap_err)
      {
        if (cap_err & 0x300)
  	  {
  	  if (mtr->fifo_errors % 50 == 0)
  	      {
  	        debug_msg ("meteor%d: capture error, %s FIFO overflow.\n", 
	          mtr->index, cap_err & 0x0100 ? "even" : "odd");
	      }
	    mtr->fifo_errors++;	/* increment fifo capture errors cnt */
  	  }
        if (cap_err & 0xc00)
	  {
	    if (mtr->dma_errors % 50 == 0)
	      {
	        debug_msg ("meteor%d: capture error, %s DMA address error.\n", 
                  mtr->index, cap_err & 0x0400 ? "even" : "odd");
	      }
	    mtr->dma_errors++;	/* increment DMA capture errors cnt */
	  }
        s7116->capt_ctl |= 0x0f00;		/* clear error */
      }

    return 0;
   }
//  wait_jiffies (WAIT_JIFFS);	/* Let board settle */

  return 1;

}

/* ==============================================================
 *
 *    D E V I C E   F U N C T I O N S
 *
 * ============================================================== */

static void
meteor_initialize(struct meteor *mtr)
{
  struct pci_dev *dev = mtr->pci;
  struct saa7116 *s7116;
  __u8  latency;
  __u16 command;

  #ifdef DEBUG 
    __u8  revision;
  #endif

  #ifdef DEBUG 
    /* Print device revision */
    pci_read_config_byte(dev, PCI_CLASS_REVISION, &revision);
    debug_msg ("<SAA 7116> (Rev %d)\n", revision);
  #endif
    
  /*  Initialize Memory on saa7116  */
  pci_read_config_word (dev, PCI_COMMAND, &command);
  command |= PCI_COMMAND_MASTER;
  command |= PCI_COMMAND_MEMORY;
  if (pci_write_config_word (dev, PCI_COMMAND, command))
    {
      err_msg ("meteor%d: PCI config failed\n", mtr->index);
      return;
    }

  pci_read_config_byte(dev, PCI_LATENCY_TIMER, &latency);
  latency = METEOR_LATENCY;
  pci_write_config_byte(dev, PCI_LATENCY_TIMER, latency);

  /* Initialize the saa7116. */
  saa7116_init(mtr);
   
  /*
   * Determine the size of the dma_end registers. The documentation
   * warn that in future SAA 7116 version more bits may be added.
   */

  s7116 = mtr->s7116;
  s7116->dma_end_e = s7116->dma_end_o = ~0;
  mtr->dma_range = (s7116->dma_end_e & s7116->dma_end_o) | 0xff;
  // debug_msg ("meteor%d: DMA range: %lx\n", mtr->index, mtr->dma_range);

  return;
}


static void
meteor_brightness (struct meteor *mtr, int x)
{
  SAA7196_WRITE (mtr, SAA7196_BRIG, (unsigned char) x);
}

static void
meteor_contrast (struct meteor *mtr, int x)
{
  SAA7196_WRITE (mtr, SAA7196_CONT, (unsigned char) x);
}

static void
meteor_saturation (struct meteor *mtr, int x)
{
  SAA7196_WRITE (mtr, SAA7196_CSAT, (unsigned char) x);
}

static void
meteor_hue (struct meteor *mtr, int x)
{
  /* V4L2 hue=128 equals Meteor hue=0 */
  x = (x + 128) % 256;
  SAA7196_WRITE (mtr, SAA7196_HUEC, (unsigned char) x);
}

static void
meteor_gain (struct meteor *mtr, int x)
{
  SAA7196_WRITE (mtr, SAA7196_CGAINR, (unsigned char) x);
}

static void
meteor_tone_controls (struct meteor *mtr)
{
  int *ctrl;
  
  ctrl = mtr->source[mtr->input].control;
  meteor_hue (mtr, ctrl[VCTRL_HUE]);
  meteor_gain (mtr, ctrl[VCTRL_GAIN]);
  meteor_saturation (mtr, ctrl[VCTRL_SATURATION]);
  meteor_contrast (mtr, ctrl[VCTRL_CONTRAST]);
  meteor_brightness (mtr, ctrl[VCTRL_BRIGHTNESS]);
}

static void interrupt_enable (struct meteor *mtr);
static void interrupt_disable (struct meteor *mtr);

static void
grabbing_enable (struct meteor *mtr, int x)
{
  __u32 capt;

  if (x)
    {        
        /* Start the capture */
	// FIXME fix all the possibile values of field
	switch (mtr->clientfmt.field)
	 {
	  case V4L2_FIELD_TOP:
	    if (mtr->streaming)
	      capt = 0x8ff2;
  	    else
	      capt = 0x8ff8;
	    break;
	  case V4L2_FIELD_BOTTOM:
      	    if (mtr->streaming)
	      capt = 0x8ff1;
  	    else
  	      capt = 0x8ff4;
	    break;
          default: // V4L2_FIELD_INTERLACED || V4L2_FIELD_ALTERNATE
	    if (mtr->streaming)
	      capt = 0x8ff3;
  	    else
	      capt = 0x8ffc;
	    break;
	  }

      mtr->s7116->capt_ctl = capt;
      // debug_msg ("meteor%d:  start_capture 0x%lx 0x%lx\n", mtr->index, 
      //    mtr->s7116->capt_ctl, mtr->s7116->ints_ctl ); 
    }
  else
    {
      /* Stop the capture */
      // debug_msg ("meteor%d:  stop_capture 0x%lx 0x%lx\n", mtr->index, 
      //                mtr->s7116->capt_ctl, mtr->s7116->ints_ctl);
      mtr->s7116->capt_ctl = 0x8ff0;

      /* Let board settle */
      wait_jiffies (WAIT_JIFFS);
    }
}

static __u32
current_time_ms (void)
{
  #ifdef METEOR_USE_TIMESPEC
    struct timespec now;
    now = current_kernel_time();
    return now.tv_sec * 1000 + now.tv_nsec / 1000000;
  #else
    struct timeval now;
    do_gettimeofday (&now);
    return now.tv_sec * 1000 + now.tv_usec / 1000;
  #endif
}


/* ==============================================================
 *
 *      M E M O R Y   A L L O C A T I O N
 *
 * ============================================================== */
 
static void
himem_search (void)
{
  __u32 *himem_kernaddr, himem_size;
  __u8 *tst_addr;
  __u8 twr, trd;


  if (himem_head != NULL)
    return;
    
  /* Allocate High Memory for the Frame Buffer */
  himem_kernaddr = (__u32 *)high_memory;
  
  /* Set characteristics in the last himem_node */
  himem_head = (struct himem_node*)vmalloc(sizeof(struct himem_node));
  himem_head->addr = (__u32)virt_to_phys( (void *)himem_kernaddr );
  
  himem_size = 0;
  for(twr=1; himem_size < (32<<20); twr++) 
   { // himem_size tested to prevent inf loop
    tst_addr = ioremap( (__u32)(himem_head->addr + himem_size), PAGE_SIZE);
    writeb( twr, tst_addr);
    trd = twr - 1;
    trd = readb( tst_addr);
    iounmap( tst_addr );
    if( trd == twr ) 
      himem_size += PAGE_SIZE;
    else 
      break;
   }  

	       
  himem_head->len = himem_size;
  himem_head->used = 0;
  himem_head->next = NULL;
  himem_head->previous = NULL;
  
  if (himem_size == 0)
    err_msg("No free high memory, please read documentation.\n");
  else
   debug_msg("Found %dMb in higher memory at 0x%lx\n", himem_size>>20, 
               (unsigned long)himem_head->addr);
}

static __u8*
malloc_himem (struct meteor *mtr,__u32 size,struct himem_node **fbuf)
{
  struct himem_node *tmp,*last;
  __u32 buf_size, himem_size, block_size_left;
  __u8 *remap;
  unsigned long rpage;
  unsigned long roffset;
  
  if (himem_head == NULL)
    return NULL;
    
  buf_size = PAGE_ALIGN (size);

  /* Find last node containing info on the resting size of mem */
  for (tmp = himem_head; tmp->next != NULL; tmp = tmp->next);
  himem_size = tmp->len;

  /* TODO: Always allocate new node at the end, could be improved by
           pruning free space in the himem list */
  if (buf_size > himem_size)
   {
    err_msg("Needs more free higher memory (%dkB < %dkB)\n", 
              himem_size>>10, buf_size>>10);
    return NULL;
   }
  if (buf_size > mtr->dma_range)
   {
    err_msg("size too big for DMA capability (%ldkB < %dkB)\n", 
              (mtr->dma_range+1)>>10, buf_size>>10);
    return NULL;
   }
  
  /*
   * It is important that frames do not cross 4Mb boundaries.
   * The dma_end registers only have 22 significant bits, and even
   * without range checking the dma regs themselves don't seem to
   * increment over 4Mb boundaries (they just wrap around into 
   * the current 4Mb block).
   */
  block_size_left = mtr->dma_range+1 - (mtr->dma_range & tmp->addr);
  if (buf_size > block_size_left)
   {
     buf_size += block_size_left;
     if (buf_size > himem_size)
      {
       err_msg("Needs more free higher memory (%dkB < %dkB)\n", 
              himem_size>>10, buf_size>>10);
       return NULL;
      }
     tmp->addr += block_size_left;
   }
         
  /* Allocate a new node where to set the memory */
  last = (struct himem_node*)vmalloc(sizeof(struct himem_node));
  last->addr = tmp->addr + buf_size;
  last->len = himem_size - buf_size;
  last->used = 0;
  last->next = NULL;
  last->previous = tmp; 
  
  tmp->len = buf_size;
  tmp->used = 1;
  tmp->next = last;
  
  *fbuf = tmp;

  /* TODO: Useless precaution ? */
  rpage = (unsigned long)tmp->addr & PAGE_MASK;
  roffset = (unsigned long)tmp->addr & ~PAGE_MASK;
  
  remap = (__u8 *)(roffset | (u_long)ioremap(rpage, buf_size));
  // debug_msg("Allocated %dkB of High Memory at %lx\n", buf_size>>10, rpage);
  
  return remap;
}

static void
free_himem (__u8 *buffer, struct himem_node *fbuf)
{
  struct himem_node *tmp;
  
  if ((buffer == NULL) || (fbuf == NULL))
    return;
  
  if (fbuf->used == 0)
    return;
    
  fbuf->used = 0;
  if (fbuf->next->used == 0)
   {
    /* Case next is unused: group them in fbuf delete next */
    tmp = fbuf->next;
    fbuf->len = fbuf->len + tmp->len;
    fbuf->next = tmp->next;
    if (tmp->next != NULL)
      tmp->next->previous = fbuf;
    vfree(tmp);
   }
  if (fbuf->previous != NULL)
   {
    /* Case previous is unused: group them in previous delete fbuf */
    tmp = fbuf->previous;
    if (tmp->used == 0)
     {
      tmp->len = tmp->len + fbuf->len;
      tmp->next = fbuf->next;
      if (fbuf->next != NULL)
        fbuf->next->previous = tmp;
      vfree(fbuf);
     }
   }
   
  /* Unmap buffer */
  iounmap(buffer);
  buffer = NULL;
  
  return;
}

/* ==============================================================
 *
 *      S A A 7 1 9 6   &   B T 2 5 4   V I D E O   D E C O D E R S
 *
 * ============================================================== */

/* Switch from SAA7196 to BT254 */

static void
select_bt254 (struct meteor *mtr)
{

  /* disable saa7196, saaen = 1 */
  PCF8574_CTRL_WRITE (mtr, PCF8574_CTRL_REG (mtr) | 0x80);
  /* enable Bt254, bten = 0 */
  PCF8574_CTRL_WRITE (mtr, PCF8574_CTRL_REG (mtr) & ~0x40);
}

/* Switch from Bt254 to SAA7196 */

static void
select_saa7196 (struct meteor *mtr)
{
  /* disable Bt254, bten = 1 */
  PCF8574_CTRL_WRITE (mtr, PCF8574_CTRL_REG (mtr) | 0x40);
  /* enable saa7196, saaen = 0 */
  PCF8574_CTRL_WRITE (mtr, PCF8574_CTRL_REG (mtr) & ~0x80);
}

/* BT 254 Procedures */

static void
bt254_write (struct meteor *mtr, __u8 addr, __u8 data)
{
  addr &= 0x7;			/* sanity? */
  mtr->bt254_reg[addr] = data;
  PCF8574_DATA_WRITE (mtr, data);	/* set data */
  PCF8574_CTRL_WRITE (mtr, (PCF8574_CTRL_REG (mtr) & ~0x7) | addr);
  PCF8574_CTRL_WRITE (mtr, PCF8574_CTRL_REG (mtr) & ~0x10);	/* WR/ to 0 */
  PCF8574_CTRL_WRITE (mtr, PCF8574_CTRL_REG (mtr) | 0x10);	/* WR to 1 */
  PCF8574_DATA_WRITE (mtr, 0xff);	/* clr data */

}

/* SAA 7196 Procedures */

static int
p7196_check (struct meteor *mtr)
{
 SAA7196_WRITE (mtr, SAA7196_STDC,
 	     SAA7196_REG (mtr, SAA7196_STDC) & ~0x02);
 SAA7196_READ (mtr);
 
 // debug_msg ("meteor%d: SAA 7196 Status Byte 0 0x%x\n",
 //	 mtr->index, mtr->s7116->i2c_stat.r_data);
		 
 SAA7196_WRITE (mtr, SAA7196_STDC,
	     SAA7196_REG (mtr, SAA7196_STDC) | 0x02);
 SAA7196_READ (mtr);
 
 // debug_msg ("meteor%d: SAA 7196 Status Byte 1 0x%x\n",
 // 	 mtr->index, mtr->s7116->i2c_stat.r_data);
	 
  return 1;
}

static int
p7196_initialize (struct meteor *mtr)
{
  /*  Video decoder information fields */
  mtr->videc.standards = V4L2_STD_NTSC | V4L2_STD_PAL | V4L2_STD_SECAM;
  /*
  mtr->videc.standards = (1<<V4L2_STD_NTSC_SHIFT) 
	  | (1<<V4L2_STD_PAL_SHIFT) 
	  | (1<<V4L2_STD_SECAM_SHIFT) ;
  */
  mtr->videc.ntsc_hskip = 30;
  mtr->videc.ntsc_vskip = 12;
  mtr->videc.ntsc_width = 640;
  mtr->videc.ntsc_height = 480;
  mtr->videc.ntsc_field_order = 0;
  mtr->videc.pal_hskip = 62;
  mtr->videc.pal_vskip = 14;
  mtr->videc.pal_width = 768;
  mtr->videc.pal_height = 576;
  mtr->videc.pal_field_order = 0;
  mtr->videc.preferred_field = 0;

  if (mtr->flags & METEOR_RGB)
    mtr->videc.num_inputs = 6;
  else
    mtr->videc.num_inputs = 5;

  /* Disable VRAM output during setup (Re-Enabled in set_oformat) 
  SAA7196_WRITE (mtr, SAA7196_FMTS, SAA7196_REG (mtr, SAA7196_FMTS) & ~0x10);
  */
  
  return 1;
}

static int
p7196_set_oformat (struct meteor *mtr)
{
  __u32 buffer;
  int size, row_width, cr_cols, cr_rows;
  register struct saa7116 *s7116 = mtr->s7116;

  /* set defaults and end of buffer locations */
  mtr->dma_add_e[0] = 0;	/* DMA 1 even    */
  mtr->dma_add_e[1] = 0;	/* DMA 2 even    */
  mtr->dma_add_e[2] = 0;	/* DMA 3 even    */
  mtr->dma_add_o[0] = 0;	/* DMA 1 odd     */
  mtr->dma_add_o[1] = 0;	/* DMA 2 odd     */
  mtr->dma_add_o[2] = 0;	/* DMA 3 odd     */
  mtr->dma_str_e[0] = 0;	/* Stride 1 even */
  mtr->dma_str_e[1] = 0;	/* Stride 2 even */
  mtr->dma_str_e[2] = 0;	/* Stride 3 even */
  mtr->dma_str_o[0] = 0;	/* Stride 1 odd  */
  mtr->dma_str_o[1] = 0;	/* Stride 2 odd  */
  mtr->dma_str_o[2] = 0;	/* Stride 3 odd  */

  switch (mtr->clientfmt.pixelformat)
    {
    /* YUV 9 Planar*/
    case V4L2_PIX_FMT_YUV410: 
      row_width = mtr->clientfmt.width * mtr->clientfmt.height;
      cr_cols = mtr->clientfmt.width >> 2;
      cr_rows = mtr->clientfmt.height >> 2;
      switch (mtr->clientfmt.field)
	{
        case V4L2_FIELD_ALTERNATE:
          mtr->dma_add_o[1] = row_width;               
          mtr->dma_add_o[2] = row_width + cr_cols * cr_rows;
          mtr->dma_add_e[1] = row_width;              
          mtr->dma_add_e[2] = row_width + cr_cols * cr_rows;
          SAA7196_WRITE (mtr, SAA7196_FMTS, 0xb1);
 	  break;
	case V4L2_FIELD_TOP:
	  mtr->dma_add_o[1] = row_width;
	  mtr->dma_add_o[2] = row_width + cr_cols * cr_rows;
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0xd1);
 	  break;
	case V4L2_FIELD_BOTTOM:
	  mtr->dma_add_e[1] = row_width;
	  mtr->dma_add_e[2] = row_width + cr_cols * cr_rows;
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0xf1);
 	  break;
        case V4L2_FIELD_INTERLACED:
	default:      /* (mc) FIXME the other field possibilities */
	  mtr->dma_add_o[0] = mtr->clientfmt.width;
	  mtr->dma_add_e[1] = row_width;
	  mtr->dma_add_o[1] = row_width + cr_cols;
	  mtr->dma_add_e[2] = row_width + cr_cols * cr_rows;
	  mtr->dma_add_o[2] = row_width + cr_cols * cr_rows + cr_cols;
	  mtr->dma_str_e[0] = mtr->dma_str_o[0] = mtr->clientfmt.width;
	  mtr->dma_str_e[1] = mtr->dma_str_o[1] = cr_cols;
	  mtr->dma_str_e[2] = mtr->dma_str_o[2] = cr_cols;
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0x91);
	  break;
	}
      s7116->rt_mode_e = s7116->rt_mode_o =(struct rt_mode){mode: 0xc3, route:0xaaaaff};
      break;
    /* YUV 12 Planar*/
    case V4L2_PIX_FMT_YUV420:
      row_width = mtr->clientfmt.width * mtr->clientfmt.height;
      cr_cols = mtr->clientfmt.width >> 1;
      cr_rows = mtr->clientfmt.height >> 1;
      switch (mtr->clientfmt.field)
	{
        case V4L2_FIELD_ALTERNATE:
          mtr->dma_add_o[1] = row_width;               
          mtr->dma_add_o[2] = row_width + cr_cols * cr_rows;
          mtr->dma_add_e[1] = row_width;              
          mtr->dma_add_e[2] = row_width + cr_cols * cr_rows;
          SAA7196_WRITE (mtr, SAA7196_FMTS, 0xb1);
 	  break;
	case V4L2_FIELD_TOP:
	  mtr->dma_add_o[1] = row_width;
	  mtr->dma_add_o[2] = row_width + cr_cols * cr_rows;
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0xd1);
 	  break;
	case V4L2_FIELD_BOTTOM:
	  mtr->dma_add_e[1] = row_width;
	  mtr->dma_add_e[2] = row_width + cr_cols * cr_rows;
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0xf1);
 	  break;
        case V4L2_FIELD_INTERLACED:
	default:                         // (mc) FIXME
	  mtr->dma_add_o[0] = mtr->clientfmt.width;
	  mtr->dma_add_e[1] = row_width;
	  mtr->dma_add_o[1] = row_width + cr_cols;
	  mtr->dma_add_e[2] = row_width + cr_cols * cr_rows;
	  mtr->dma_add_o[2] = row_width + cr_cols * cr_rows + cr_cols;
	  mtr->dma_str_e[0] = mtr->dma_str_o[0] = mtr->clientfmt.width;
	  mtr->dma_str_e[1] = mtr->dma_str_o[1] = cr_cols;
	  mtr->dma_str_e[2] = mtr->dma_str_o[2] = cr_cols;
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0x91);
	  break;
	}
      s7116->rt_mode_e = s7116->rt_mode_o =(struct rt_mode){mode: 0xc2, route:0xaaaaff};
      break;
    /* YUV 16 Planar*/
    case V4L2_PIX_FMT_YYUV:
      row_width = mtr->clientfmt.width * mtr->clientfmt.height;
      cr_cols = mtr->clientfmt.width >> 1;
      cr_rows = mtr->clientfmt.height;
      switch (mtr->clientfmt.field)
	{
        case V4L2_FIELD_ALTERNATE:
          mtr->dma_add_o[1] = row_width;               
          mtr->dma_add_o[2] = row_width + cr_cols * cr_rows;
          mtr->dma_add_e[1] = row_width;              
          mtr->dma_add_e[2] = row_width + cr_cols * cr_rows;
          SAA7196_WRITE (mtr, SAA7196_FMTS, 0xb1);
 	  break;
	case V4L2_FIELD_TOP:
	  mtr->dma_add_o[1] = row_width;
	  mtr->dma_add_o[2] = row_width + cr_cols * cr_rows;
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0xd1);
 	  break;
	case V4L2_FIELD_BOTTOM:
	  mtr->dma_add_e[1] = row_width;
	  mtr->dma_add_e[2] = row_width + cr_cols * cr_rows;
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0xf1);
 	  break;
        case V4L2_FIELD_INTERLACED:
	default:                         // (mc) FIXME
	  mtr->dma_add_o[0] = mtr->clientfmt.width;
	  mtr->dma_add_e[1] = row_width;
	  mtr->dma_add_o[1] = row_width + cr_cols;
	  mtr->dma_add_e[2] = row_width + cr_cols * cr_rows;
	  mtr->dma_add_o[2] = row_width + cr_cols * cr_rows + cr_cols;
	  mtr->dma_str_e[0] = mtr->dma_str_o[0] = mtr->clientfmt.width;
	  mtr->dma_str_e[1] = mtr->dma_str_o[1] = cr_cols;
	  mtr->dma_str_e[2] = mtr->dma_str_o[2] = cr_cols;
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0x91);
	  break;
	}
      s7116->rt_mode_e = s7116->rt_mode_o =(struct rt_mode){mode: 0xc1, route:0xaaaaff};
      break;
    /* YUV 16 Packed*/
    case V4L2_PIX_FMT_UYVY:
      switch (mtr->clientfmt.field)
	{
        case V4L2_FIELD_ALTERNATE:
          SAA7196_WRITE (mtr, SAA7196_FMTS, 0xb1);
 	  break;
	case V4L2_FIELD_TOP:
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0xd1);
 	  break;
	case V4L2_FIELD_BOTTOM:
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0xf1);
 	  break;
        case V4L2_FIELD_INTERLACED:
	default:                         // (mc) FIXME
	  /* 
	   * Be carefull to use capture_bypp instead of clientfmt.depth 
	   * here to take into consideration the fact there could be a
	   * a translation. Same comments for later use in this function.
	   */
	  mtr->dma_add_o[0] = mtr->dma_str_e[0] = mtr->dma_str_o[0] =
	    (mtr->clientfmt.width * mtr->capture_bypp);
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0x91);
	  break;
	}
      s7116->rt_mode_e = s7116->rt_mode_o =(struct rt_mode){mode: 0x41, route:0xeeeeee};
      break;
    /* RGB 16 Packed*/
    case V4L2_PIX_FMT_RGB555:
     switch (mtr->clientfmt.field)
	{
        case V4L2_FIELD_ALTERNATE:
          SAA7196_WRITE (mtr, SAA7196_FMTS, 0xb0);
 	  break;
	case V4L2_FIELD_TOP:
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0xd0);
 	  break;
	case V4L2_FIELD_BOTTOM:
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0xf0);
 	  break;
        case V4L2_FIELD_INTERLACED:
	default:                         // (mc) FIXME
	  mtr->dma_add_o[0] = mtr->dma_str_e[0] = mtr->dma_str_o[0] =
	    (mtr->clientfmt.width * mtr->capture_bypp);
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0x90);
	  break;
	}
      s7116->rt_mode_e = s7116->rt_mode_o =(struct rt_mode){mode: 0x01, route:0xeeeeee};
      break;
    /* RGB 32 Packed*/
    case V4L2_PIX_FMT_BGR24:
    case V4L2_PIX_FMT_BGR32:
      switch (mtr->clientfmt.field)
	{
        case V4L2_FIELD_ALTERNATE:
          SAA7196_WRITE (mtr, SAA7196_FMTS, 0xb2);
 	  break;
	case V4L2_FIELD_TOP:
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0xd2);
 	  break;
	case V4L2_FIELD_BOTTOM:
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0xf2);
 	  break;
        case V4L2_FIELD_INTERLACED:
	default:                         // (mc) FIXME
	  mtr->dma_add_o[0] = mtr->dma_str_e[0] = mtr->dma_str_o[0] =
	    (mtr->clientfmt.width * mtr->capture_bypp);
	  SAA7196_WRITE (mtr, SAA7196_FMTS, 0x92);
	  break;
	}
      s7116->rt_mode_e = s7116->rt_mode_o =(struct rt_mode){mode: 0x00, route:0x393939};
      break;
    default:
      debug_msg ("unknown pixel format %4.4s\n",
		 (char *) &mtr->clientfmt.pixelformat);
      return 0;
      break;
    }

  /* Acquire a first field */
  if (mtr->capture_himem_node != NULL)
   {
     buffer = mtr->capture_himem_node->addr;
     size = mtr->capture_himem_node->len;
   }
  else if (mtr->stream_current_buffer != NULL)
   {
     buffer = mtr->stream_current_buffer->himem_node->addr;
     size = mtr->stream_current_buffer->himem_node->len;
   } 
  else
   {
     err_msg("capture buffer not set\n");
     return 0;
   } 
      
  s7116->dma_e[0] = mtr->dma_add_e[0] + buffer;
  s7116->dma_e[1] = mtr->dma_add_e[1] + buffer;
  s7116->dma_e[2] = mtr->dma_add_e[2] + buffer;
  s7116->dma_o[0] = mtr->dma_add_o[0] + buffer;
  s7116->dma_o[1] = mtr->dma_add_o[1] + buffer;
  s7116->dma_o[2] = mtr->dma_add_o[2] + buffer;
  s7116->stride_e[0] = mtr->dma_str_e[0];
  s7116->stride_e[1] = mtr->dma_str_e[1];
  s7116->stride_e[2] = mtr->dma_str_e[2];
  s7116->stride_o[0] = mtr->dma_str_o[0];
  s7116->stride_o[1] = mtr->dma_str_o[1];
  s7116->stride_o[2] = mtr->dma_str_o[2];
  s7116->dma_end_e = s7116->dma_end_o = (buffer + size - 1);

  /* Set Output Line */
  SAA7196_WRITE (mtr, SAA7196_OUTPIX, mtr->clientfmt.width & 0xff);
  SAA7196_WRITE (mtr, SAA7196_HFILT,
		 ((SAA7196_REG (mtr, SAA7196_HFILT) & ~0x03) | ((mtr->clientfmt.width >> 8) & 0x03)));

  /* Set Output Height of one Field */
  if ((mtr->clientfmt.field) == V4L2_FIELD_INTERLACED)
   {
    SAA7196_WRITE (mtr, SAA7196_OUTLINE, (mtr->clientfmt.height >> 1) & 0xff);
    SAA7196_WRITE (mtr, SAA7196_VYP,
		 ((SAA7196_REG (mtr, SAA7196_VYP) & ~0x03) | ((mtr->clientfmt.height >> 9) & 0x03)));
   }
  else
   {
    SAA7196_WRITE (mtr, SAA7196_OUTLINE, mtr->clientfmt.height & 0xff);
    SAA7196_WRITE (mtr, SAA7196_VYP,
		 ((SAA7196_REG (mtr, SAA7196_VYP) & ~0x03) | ((mtr->clientfmt.height >> 8) & 0x03)));
   }		 
  
  return 1;
}

static int
p7196_set_input (struct meteor *mtr, int i)
{

  if (i < 0 || i >= mtr->videc.num_inputs)
    return -EINVAL;
    
  mtr->input = i;

  switch (i)
    {
    case METEOR_INPUT_DEV0:
      if (mtr->flags & METEOR_RGB)
	select_saa7196 (mtr);
      SAA7196_WRITE (mtr, SAA7196_IOCC,
		     (SAA7196_REG (mtr, SAA7196_IOCC) & ~0x3) | 0x0);
      SAA7196_WRITE (mtr, SAA7196_LUMC,
		     (SAA7196_REG (mtr, SAA7196_LUMC) & ~0x80));
      break;
    case METEOR_INPUT_DEV1:
      if (mtr->flags & METEOR_RGB)
	select_saa7196 (mtr);
      SAA7196_WRITE (mtr, SAA7196_IOCC,
		     (SAA7196_REG (mtr, SAA7196_IOCC) & ~0x3) | 0x1);
      SAA7196_WRITE (mtr, SAA7196_LUMC,
		     (SAA7196_REG (mtr, SAA7196_LUMC) & ~0x80));
      break;
    case METEOR_INPUT_DEV2:
      if (mtr->flags & METEOR_RGB)
	select_saa7196 (mtr);
      SAA7196_WRITE (mtr, SAA7196_IOCC,
		     (SAA7196_REG (mtr, SAA7196_IOCC) & ~0x3) | 0x2);
      SAA7196_WRITE (mtr, SAA7196_LUMC,
		     (SAA7196_REG (mtr, SAA7196_LUMC) & ~0x80));
      break;
    case METEOR_INPUT_DEV3:
      if (mtr->flags & METEOR_RGB)
	select_saa7196 (mtr);
      SAA7196_WRITE (mtr, SAA7196_IOCC,
		     (SAA7196_REG (mtr, SAA7196_IOCC) & ~0x3) | 0x3);
      SAA7196_WRITE (mtr, SAA7196_LUMC,
		     (SAA7196_REG (mtr, SAA7196_LUMC) & ~0x80));
      break;
    case METEOR_INPUT_DEV_SVIDEO:
      if (mtr->flags & METEOR_RGB)
	select_saa7196 (mtr);
      SAA7196_WRITE (mtr, SAA7196_IOCC,
		     (SAA7196_REG (mtr, SAA7196_IOCC) & ~0x3) | 0x2);
      SAA7196_WRITE (mtr, SAA7196_LUMC,
		     (SAA7196_REG (mtr, SAA7196_LUMC) & ~0x80) | 0x80);
      break;
    case METEOR_INPUT_DEV_RGB:
      if ((mtr->flags & METEOR_RGB) == 0)
	return (-EINVAL);
      SAA7196_WRITE (mtr, SAA7196_IOCC,
		     (SAA7196_REG (mtr, SAA7196_IOCC) & ~0x3) | 0x3);
      SAA7196_WRITE (mtr, SAA7196_LUMC,
		     (SAA7196_REG (mtr, SAA7196_LUMC) & ~0x80));
      select_bt254 (mtr);
      SAA7196_WRITE (mtr, SAA7196_IOCC,	/* chn 3 for synch */
		     (SAA7196_REG (mtr, SAA7196_IOCC) & ~0x3) | 0x3);
      break;
    default:
      return (-EINVAL);
    }
    
  mtr->videc.decoder_is_stable = mtr->videc.is_stable(mtr);

  return 1;
}

static int
p7196_set_standard (struct meteor *mtr, int id )
{
  mtr->videc.standard = id;

  /* Switch the hardware to the new standard  */
  if ( id & V4L2_STD_NTSC ) {
      SAA7196_WRITE (mtr, SAA7196_STDC,
		     (SAA7196_REG (mtr, SAA7196_STDC) & ~0x01));
      SAA7196_WRITE (mtr, SAA7196_CTRL1,
		     (SAA7196_REG (mtr, SAA7196_CTRL1) & ~0xe0) | 0x40);
      SAA7196_WRITE (mtr, SAA7196_INPIX, 0x80);
      SAA7196_WRITE (mtr, SAA7196_HFILT,
		     (SAA7196_REG (mtr, SAA7196_HFILT) & ~0x0c) | 0x08);
      SAA7196_WRITE (mtr, SAA7196_INLINE, 0xf0);
      SAA7196_WRITE (mtr, SAA7196_VYP,
		     (SAA7196_REG (mtr, SAA7196_VYP) & ~0x0c));
      if (mtr->flags & METEOR_RGB)
	{
	  PCF8574_CTRL_WRITE (mtr, PCF8574_CTRL_REG (mtr) | 0x20);
	}
      strlcpy( mtr->standard_name, "NTSC", sizeof(mtr->standard_name));
  } else if ( id & V4L2_STD_PAL ) {
      SAA7196_WRITE (mtr, SAA7196_STDC,
		     (SAA7196_REG (mtr, SAA7196_STDC) & ~0x01));
      SAA7196_WRITE (mtr, SAA7196_CTRL1,
		     (SAA7196_REG (mtr, SAA7196_CTRL1) & ~0xe0));
      SAA7196_WRITE (mtr, SAA7196_INPIX, 0x00);
      SAA7196_WRITE (mtr, SAA7196_HFILT,
		     (SAA7196_REG (mtr, SAA7196_HFILT) | 0x0c));
      SAA7196_WRITE (mtr, SAA7196_INLINE, 0x20);
      SAA7196_WRITE (mtr, SAA7196_VYP,
		     (SAA7196_REG (mtr, SAA7196_VYP) & ~0x0c) | 0x04);
      if (mtr->flags & METEOR_RGB)
	{
	  PCF8574_CTRL_WRITE (mtr, PCF8574_CTRL_REG (mtr) &= ~0x20);
	}
      strlcpy( mtr->standard_name, "PAL", sizeof(mtr->standard_name));
  } else if ( id & V4L2_STD_SECAM ) {
      SAA7196_WRITE (mtr, SAA7196_STDC,
		     (SAA7196_REG (mtr, SAA7196_STDC) & ~0x01) | 0x1);
      SAA7196_WRITE (mtr, SAA7196_CTRL1,
		     (SAA7196_REG (mtr, SAA7196_CTRL1) & ~0xe0) | 0x20);
      SAA7196_WRITE (mtr, SAA7196_INPIX, 0x00);
      SAA7196_WRITE (mtr, SAA7196_HFILT,
		     (SAA7196_REG (mtr, SAA7196_HFILT) | 0x0c));
      SAA7196_WRITE (mtr, SAA7196_INLINE, 0x20);
      SAA7196_WRITE (mtr, SAA7196_VYP,
		     (SAA7196_REG (mtr, SAA7196_VYP) & ~0x0c) | 0x04);
      if (mtr->flags & METEOR_RGB)
	{
	  PCF8574_CTRL_WRITE (mtr, PCF8574_CTRL_REG (mtr) &= ~0x20);
	}

      strlcpy( mtr->standard_name, "SECAM", sizeof(mtr->standard_name));
  } else {
      /* Autodetect PAL or NTSC (SECAM not detected) */
      SAA7196_WRITE(mtr, SAA7196_STDC, 
                (SAA7196_REG(mtr, SAA7196_STDC) & ~0x01));
      SAA7196_WRITE(mtr, SAA7196_CTRL1, 
                (SAA7196_REG(mtr, SAA7196_CTRL1) & ~0xe0) | 0x80);
      strlcpy( mtr->standard_name, "AUTO", sizeof(mtr->standard_name));
      return 0;
  }
  return 1;
}

/* TODO: VCR mode not taken into consideration in V4L2 API ? */
static int
p7196_set_vcrmode (struct meteor *mtr, int x)
{
  mtr->source[mtr->input].vcrmode = x;

  if (x)
   {
    SAA7196_WRITE (mtr, SAA7196_STDC,(SAA7196_REG (mtr, SAA7196_STDC) | 0x80));
   }
  else
   {
    SAA7196_WRITE (mtr, SAA7196_STDC,(SAA7196_REG (mtr, SAA7196_STDC) & ~0x80));
   }

  return 1;
}

static int
p7196_init_reg (struct meteor *mtr)
{
  __u32 i;
  
  /*
   * Initialize SAA7196
   */
  for (i = 0; i < NUM_SAA7196_I2C_REGS; i++)
    SAA7196_WRITE (mtr, i, saa7196_i2c_default[i]);


  /*
     * Initialize BT254 for Meteor RGB 
   */
  if (mtr->flags & METEOR_RGB)
    {
      PCF8574_CTRL_WRITE (mtr, 0x7f);
      PCF8574_DATA_WRITE (mtr, 0xff);	/* data port must be 0xff */
      PCF8574_CTRL_WRITE (mtr, 0x7f);

      /* init RGB module for 24bpp, composite input */
      for (i = 0; i < NUM_BT254_REGS; i++)
	bt254_write (mtr, i, bt254_default[i]);

      bt254_write (mtr, BT254_COMMAND, 0x00);	/* 24 bpp */
    }
    
   return 1;
}

static int
p7196_is_stable (struct meteor *mtr)
{
  __u32 status;
  register struct saa7116 *s7116 = mtr->s7116;
      
  /* 
   * Look for a video signal 
   */
  SAA7196_WRITE (mtr, SAA7196_STDC, SAA7196_REG (mtr, SAA7196_STDC) | 0x02);
  SAA7196_READ (mtr);
  status = s7116->i2c_stat.r_data;
  if ((status & 0x40) == 0)
    {
      if (status & 0x20)
	{
	  mtr->videc.frame_period = (struct v4l2_fract){1001, 30000}; // 333667;
          #ifdef DEBUG
	    if (status & 0x01)
	      debug_msg ("meteor%d: color video signal (30Hz)\n", mtr->index);
	    else
	      debug_msg ("meteor%d: B/W video signal (30Hz)\n", mtr->index);
          #endif
	}
      else if (status & 0x02)
	{
          mtr->videc.frame_period = (struct v4l2_fract){1, 25}; // 400000; 
          #ifdef DEBUG
	    if (status & 0x01)
	      debug_msg ("meteor%d: color video signal (25Hz)\n", mtr->index);
	    else
	      debug_msg ("meteor%d: B/W video signal (25Hz)\n", mtr->index);
          #endif
	}
      else 
	{
	  mtr->videc.frame_period = (struct v4l2_fract){1, 25}; // 400000;
	}
    }
  else
    {
      debug_msg ("meteor%d: no video signal\n", mtr->index);
      return 0;
    }

  /* set mtr->frame_period accordingly */
  mtr->frame_period = v4l2_fract_to_100ns( mtr->videc.frame_period );
    
  return 1;
}

static int
p7196_probe (struct meteor *mtr)
{

  /*
   * Check for the Philips SAA7196
   */

  if (saa7116_i2c(mtr, SAA7196_I2C_ADDR_W, 0, 0xff) == 0)
    {
      /*
       * Get version number.
       */
      SAA7196_WRITE (mtr, SAA7196_STDC,
		     SAA7196_REG (mtr, SAA7196_STDC) & ~0x02);
      SAA7196_READ (mtr);

      // debug_msg ("meteor%d: <SAA 7196> (Rev %d)\n",
      //	 mtr->index, (mtr->s7116->i2c_stat.r_data >> 4));
    }
  else
    return 0;			/* Not found */

  /*
   * Check for RGB module, initializes flags value.
   */
  if (saa7116_i2c(mtr,  PCF8574_CTRL_I2C_ADDR_W, 0, 0xff) == 0)
    {
      // debug_msg ("meteor%d: <BT 254> (RGB module)\n", mtr->index);
      mtr->flags = METEOR_RGB;
    }
  else
    {
      // debug_msg ("meteor%d: i2c abort normal if no RGB module\n", mtr->index);
      mtr->flags = 0;
    }

  /*  Fill in the method fields  */
  mtr->videc.initialize = p7196_initialize;
  mtr->videc.set_input = p7196_set_input;
  mtr->videc.set_standard = p7196_set_standard;
  mtr->videc.set_vcrmode = p7196_set_vcrmode;
  mtr->videc.set_oformat = p7196_set_oformat;
  mtr->videc.is_stable = p7196_is_stable;
  mtr->videc.init_reg = p7196_init_reg;
  mtr->videc.check = p7196_check;

  return 1;			/*  Found  */
}


/* ==============================================================
 *
 *	I M A G E   F O R M A T   T R A N S L A T I O N
 *
 * ============================================================== */

static int
translate_rgb32_rgb24(struct translation *xlat)
{
  __u8	*esi, *edi;
  __u32	eax, ebx, ecx;
  __u32	a, b, c, d;
  int	height = xlat->height;
  int	width = (xlat->width >> 2);
  int	row;

  esi = xlat->in;
  edi = xlat->out;
  eax = xlat->in_stride - xlat->width * 4;
  ebx = xlat->out_stride - xlat->width * 3;

  for (row = height; row; esi += eax, edi += ebx, --row)
    {
      for (ecx = width ; ecx; --ecx)
        {
          a = ((__u32 *)esi)[0];
          b = ((__u32 *)esi)[1];
          c = ((__u32 *)esi)[2];
          d = ((__u32 *)esi)[3];
          ((__u32 *)edi)[0] =
            ((a & 0x00FFFFFF)      ) |
            ((b & 0x000000FF) << 24);
          ((__u32 *)edi)[1] =
            ((b & 0x00FFFF00) >>  8) |
            ((c & 0x0000FFFF) << 16);
          ((__u32 *)edi)[2] =
            ((c & 0x00FF0000) >> 16) |
            ((d & 0x00FFFFFF) <<  8);
          esi += 16;
          edi += 12;
        }
    }
  return 1;
}

static void
translate_close(struct meteor *mtr)
{
	mtr->translation.type = XLAT_NULL;
	mtr->translation.in = NULL;
	mtr->translation.out = NULL;

	if (mtr->xlat_temp)
		vfree(mtr->xlat_temp);
	mtr->xlat_temp = NULL;
}

static int
translate_setup(struct meteor *mtr)
{
	int	npix2;
	translate_close(mtr);

	/*  Translation: BGR32 to BGR24 */
	mtr->translation.width = mtr->clientfmt.width;
	mtr->translation.height = mtr->clientfmt.height;
	mtr->translation.in_stride = mtr->translation.width * 4;
	mtr->translation.output_size = mtr->clientfmt.sizeimage;

	npix2 = mtr->translation.width * mtr->translation.height;
	switch (mtr->clientfmt.pixelformat)
	{
	case V4L2_PIX_FMT_BGR24:
	        if (mtr->streaming)
		 {
		  err_msg("No stream capture with BGR24\n");
		  return 0;
		 }
		mtr->translation.type = XLAT_BGR32_TO_BGR24;
		mtr->translation.out_stride = mtr->translation.width * 3;
		break;
	default:
		mtr->translation.type = XLAT_NULL;
		break;
	}
	return 1;
}

static void
translate_inandout(struct meteor *mtr,
		   __u8 *input_buffer,
		   __u8 *output_buffer,
		   __u8 output_is_user_space)
{
	/*  Translation: RGB to client format */
	mtr->translation.in = input_buffer;
	mtr->translation.out = output_buffer;
	mtr->translation.output_is_user = output_is_user_space;
}

static int /* length of output image or negative error */
translate_image(struct meteor *mtr,
		__u8	*input_buffer,
		__u8	*output_buffer,
		int	len,
		int	output_is_user)
{
	int	err;

	/* The buffer must be large enough for the whole image */
	if (len < mtr->translation.output_size)
	{
		debug_msg("Read buffer too small, %d < %d\n",
			  len, mtr->translation.output_size);
		return -EFAULT;
	}
	if (len > mtr->translation.output_size)
		len = mtr->translation.output_size;

	translate_inandout(mtr, input_buffer, output_buffer, output_is_user);

	if (mtr->translation.type == XLAT_NULL)
	{
		if (mtr->translation.in == mtr->translation.out)
			return len;
                /* DMA copy directly in the buffer so no need to do memcpy */
		if (output_is_user)
		{
			err = copy_to_user(output_buffer, 
					   mtr->translation.in, len);
			len = (err) ? -EFAULT : len;
		}
		return len;
	}

	if (output_is_user && !access_ok(VERIFY_WRITE, output_buffer, len))
	{
		debug_msg("Buffer verify failed in translate_image\n");
		return -EFAULT;
	}

	switch (mtr->translation.type)
	{
	case XLAT_BGR32_TO_BGR24:
		translate_rgb32_rgb24(&mtr->translation);
		break;
	default:
		debug_msg("Unknown image translation\n");
		break;
	}
	mtr->translation.out = NULL;
	return len;
}


/* ==============================================================
 *
 *      V I D E O   C A P T U R E   F U N C T I O N S
 *
 * ============================================================== */

/*  The image format has changed, width, height, pixel format.
 *  Decide if the format is ok or take the closest valid format.
 */
static int
capture_new_format (struct meteor *mtr)
{
  int ntsc;
  int depth;

  ntsc = ( (mtr->videc.standard & V4L2_STD_NTSC) != 0 );
  mtr->ready_to_capture = 0;

  mtr->source_width =
    (ntsc) ? mtr->videc.ntsc_width : mtr->videc.pal_width;
  mtr->source_height =
    (ntsc) ? mtr->videc.ntsc_height : mtr->videc.pal_height;


  switch (mtr->clientfmt.pixelformat)
    {
    case V4L2_PIX_FMT_YUV410:
      depth = 9;
      mtr->capture_bypp = 2;
      break;
    case V4L2_PIX_FMT_YUV420:
      depth = 12;
      mtr->capture_bypp = 2;
      break;
    case V4L2_PIX_FMT_YYUV:
      depth = 16;
      mtr->capture_bypp = 2;
      break;
    case V4L2_PIX_FMT_UYVY:
      depth = 16;
      mtr->capture_bypp = 2;
      break;
    case V4L2_PIX_FMT_RGB555:
      depth = 16;
      mtr->capture_bypp = 2;
      break;
    case V4L2_PIX_FMT_BGR24:
      depth = 24;
      mtr->capture_bypp = 4;
      break;
    case V4L2_PIX_FMT_BGR32:
      depth = 32;
      mtr->capture_bypp = 4;
      break;
    default:
      err_msg ("unknown format %4.4s\n",
		 (char *) &mtr->clientfmt.pixelformat);
      return 0;
      break;
    }

  if (mtr->clientfmt.width < MIN_WIDTH)
    mtr->clientfmt.width = MIN_WIDTH;
  if (mtr->clientfmt.height < MIN_HEIGHT)
    mtr->clientfmt.height = MIN_HEIGHT;
  if (mtr->clientfmt.width > MAX_WIDTH)
    mtr->clientfmt.width = MAX_WIDTH;
  if (mtr->clientfmt.height > MAX_HEIGHT)
    mtr->clientfmt.height = MAX_HEIGHT;
  if (mtr->clientfmt.width > mtr->source_width)
    mtr->clientfmt.width = mtr->source_width;
  if (mtr->clientfmt.height > mtr->source_height)
    mtr->clientfmt.height = mtr->source_height;

  /*
    The dma address increments in 4 byte steps, which
    means dma_end will only be correct if columns%4 == 0. 
    For YUV we require columns%8 == 0 since for each 8 bits
    of Y there are 4 bits each for U and V.  This bit of 
    code takes the conservative option.  In field mode mode
    rows don't need to be even, but we're conservative here
    again. 
  */
  if ((mtr->clientfmt.width & 0x3f8) != mtr->clientfmt.width) 
    {
      err_msg("meteor%d: width too large or not diviside by 8 (%d)\n", 
		      mtr->index, mtr->clientfmt.width);
      return 0;
    }
  // NOTE no more COMPRESSED etc.
  if ( (mtr->clientfmt.height & 0x7fe) != mtr->clientfmt.height  
	  ||
        // ((mtr->clientfmt.field & V4L2_FMT_FLAG_FIELD_field) &&
        ((mtr->clientfmt.height & 0x3fe) != mtr->clientfmt.height) ) 
    {
      err_msg("meteor%d: height too large or not even (%d)\n", 
		      mtr->index, mtr->clientfmt.height);
      return 0;
    }

  mtr->clientfmt.sizeimage = ( mtr->clientfmt.width
			   * mtr->clientfmt.height
			   * depth ) >> 3;

  mtr->capture_size = mtr->clientfmt.width
    * mtr->clientfmt.height
    * mtr->capture_bypp;
  
  return 1;
}

static void
capture_abort (struct meteor *mtr)
{
  mtr->grabber_enabled = 0;
  /*  Turn off the capture hardware  */
  interrupt_disable(mtr);
  grabbing_enable (mtr, 0);
}

/*  Allocate buffers, and get everything ready to capture
 *  an image, but don't start capturing yet.
 */
static int
capture_begin (struct meteor *mtr)
{

  capture_abort (mtr);
  
  if (mtr->ready_to_capture)
    return mtr->ready_to_capture;

  if (mtr->capture_buffer_size < mtr->capture_size) 
    {
      /* TODO: Should also reallocate all stream buffers */
      if (mtr->capture_buffer != NULL)
	free_himem (mtr->capture_buffer, mtr->capture_himem_node);
      mtr->capture_buffer_size =
	(mtr->capture_size + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1);

      mtr->capture_buffer = 
        malloc_himem (mtr, mtr->capture_buffer_size, &(mtr->capture_himem_node));
      if (mtr->capture_buffer == NULL)
	{
	  mtr->capture_buffer_size = 0;
	  err_msg ("meteor%d: Can't allocate capture buffer"
		   " %d bytes\n", mtr->index, mtr->capture_size);
	  return mtr->ready_to_capture;
	}
    }
      
  /* Set output format now that buffered memory is allocated */
  if (!mtr->videc.set_oformat(mtr))
    return mtr->ready_to_capture;

  if (!translate_setup(mtr))
    {
        debug_msg ("Translate setup error!\n");
	return mtr->ready_to_capture;
    }

  return (mtr->ready_to_capture = 1);
}

/** Start an image capture
 */
static void
capture_grab_frame (struct meteor *mtr)
{
  __u32 buffer;
  int size;
  register struct saa7116 *s7116 = mtr->s7116;

  if (mtr->ready_to_capture && mtr->grabber_enabled)
    return;
	 
  capture_begin (mtr);
  if (!mtr->ready_to_capture)
    return;
 
  buffer = mtr->capture_himem_node->addr;
  size = mtr->capture_himem_node->len;
  
  /*  Set up stream_current_buffer to point to the buffer to  */
  /*  capture the next frame into  */
  if (mtr->streaming)
    {
     struct stream_buffer *buf;

     #ifdef METEOR_USE_LIST
       buf = meteor_list_peek_head( &mtr->stream_q_capture );
     #else
       buf = v4l2_q_peek_head (&mtr->stream_q_capture);
     #endif
     if (buf != NULL)
       {
         mtr->stream_current_buffer = buf;
         buffer = buf->himem_node->addr;
         size = buf->himem_node->len;
       }
     else 
      return;
    }
  
  saa7116_check(mtr);

  wait_jiffies(WAIT_JIFFS); /* Let board settle */

  s7116->dma_e[0] = mtr->dma_add_e[0] + buffer;
  s7116->dma_e[1] = mtr->dma_add_e[1] + buffer;
  s7116->dma_e[2] = mtr->dma_add_e[2] + buffer;
  s7116->dma_o[0] = mtr->dma_add_o[0] + buffer;
  s7116->dma_o[1] = mtr->dma_add_o[1] + buffer;
  s7116->dma_o[2] = mtr->dma_add_o[2] + buffer;
  s7116->stride_e[0] = mtr->dma_str_e[0];
  s7116->stride_e[1] = mtr->dma_str_e[1];
  s7116->stride_e[2] = mtr->dma_str_e[2];
  s7116->stride_o[0] = mtr->dma_str_o[0];
  s7116->stride_o[1] = mtr->dma_str_o[1];
  s7116->stride_o[2] = mtr->dma_str_o[2];
  s7116->dma_end_e = s7116->dma_end_o = 
        (buffer + size - 1);

#ifdef DEBUG
  debug_msg ("Allocated Adress A:0x%lx S:0x%lx/%x\n", s7116->dma_e[0], s7116->dma_end_e, size);
  debug_msg ("Offset %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx\n", 
         mtr->dma_add_e[0], mtr->dma_add_e[1], mtr->dma_add_e[2],
         mtr->dma_add_o[0], mtr->dma_add_o[1], mtr->dma_add_o[2],
         mtr->dma_str_e[0], mtr->dma_str_e[1], mtr->dma_str_e[2], 
         mtr->dma_str_o[0], mtr->dma_str_o[1], mtr->dma_str_o[2]);
#endif


  /*  Start the hardware */
  grabbing_enable (mtr, 1);
  interrupt_enable (mtr);


  mtr->grabber_enabled = 1;
  mtr->capture_completed = 0;
}

/*
 *    STREAMING CAPTURE
 */

static int			/* 1 = success; 0 = failed */
capture_queuebuffer (struct meteor *mtr,
		     struct v4l2_buffer *vidbuf)
{
  int i = vidbuf->index;
  struct stream_buffer *buf = NULL;

  if (!mtr->stream_buffers_mapped)
    {
      debug_msg ("QBUF no buffers mapped\n");
      return 0;
    }
  if (vidbuf->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
    {
      debug_msg ("QBUF wrong type\n");
      return 0;
    }
  if (i < 0 || i >= MAX_NUM_FRAMES || !mtr->stream_buf[i].requested)
    {
      debug_msg ("QBUF buffer index %d is out of range\n", i);
      return 0;
    }

  buf = &mtr->stream_buf[i];

  if (!(buf->vidbuf.flags & V4L2_BUF_FLAG_MAPPED))
    {
      debug_msg ("QBUF buffer %d is not mapped\n", i);
      return 0;
    }
  if ((buf->vidbuf.flags & V4L2_BUF_FLAG_QUEUED))
    {
      debug_msg ("QBUF buffer %d is already queued\n", i);
      return 0;
    }

  buf->vidbuf.flags &= ~V4L2_BUF_FLAG_DONE;
  #ifdef METEOR_USE_LIST
    meteor_list_add_tail( &buf->qnode, &mtr->stream_q_capture );
  #else
    v4l2_q_add_tail (&mtr->stream_q_capture, &buf->qnode);
  #endif
  buf->vidbuf.flags |= V4L2_BUF_FLAG_QUEUED;
  return 1;
}

static int			/* 1 = got a buffer; 0 = no buffers */
capture_dequeuebuffer (struct meteor *mtr,
		       struct v4l2_buffer *buf)
{
  struct stream_buffer *newbuf;

  if (buf->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
    {
      debug_msg ("DQBUF wrong buffer type\n");
      return 0;
    }
  #ifdef METEOR_USE_LIST
    newbuf = meteor_list_pop_head( &mtr->stream_q_done );
  #else
    newbuf = v4l2_q_del_head (&mtr->stream_q_done);
  #endif

  if (newbuf == NULL)
  {
      debug_msg ("DQBUF nothing on done queue\n");
      return 0;
  }
  
  newbuf->vidbuf.flags &= ~V4L2_BUF_FLAG_QUEUED;

  *buf = newbuf->vidbuf;
  return 1;
}

static int
capture_streamon (struct meteor *mtr,
		  __u32 type)
{
  #ifndef METEOR_USE_LIST
    struct stream_buffer *buf;
  #endif

  if (type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
    {
      debug_msg ("STREAMON wrong buffer type\n");
      return 0;
    }
  if (mtr->streaming)
    {
      return 1;
    }

  capture_abort (mtr);	/* cancel any capture that might be in progress */

  /*  -2 is a magic number that triggers start-of-stream logic in */
  /*    capture_interrupt()  */
  mtr->stream_last_frame = -2;

#ifdef CONFIG_V4L2_PERFORMANCE
  mtr->perf.frames = 0;
  mtr->perf.framesdropped = 0;
  mtr->perf.bytesout = 0;
#endif // CONFIG_V4L2_PERFORMANCE
  
  /*  Look for video input and set frame period accordingly */
  if (!mtr->videc.is_stable(mtr))
     return -EIO;

  //  Can't capture frames faster than the video input  
  if ( v4l2_fract_isless( mtr->capture.timeperframe, mtr->videc.frame_period ) )
    mtr->capture.timeperframe = mtr->videc.frame_period;
  mtr->timeperframe = v4l2_fract_to_100ns( mtr->capture.timeperframe );
  
  /*  Move any leftover DONE buffers to the free pool */
  #ifdef METEOR_USE_LIST
    meteor_list_unqueue( &mtr->stream_q_done );
  #else
    while ((buf = v4l2_q_del_head (&mtr->stream_q_done)))
      buf->vidbuf.flags &= ~V4L2_BUF_FLAG_QUEUED;
  #endif

  /*  Kick off the machine */
  mtr->streaming = 1;

  capture_grab_frame (mtr);

  return 1;
}

static void
capture_streamoff (struct meteor *mtr,
		   __u32 type)
{
  if (!mtr->streaming)
    return;
  if (type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
    {
      debug_msg ("STREAMOFF wrong buffer type\n");
      return;
    }
  capture_abort (mtr);
  
  mtr->stream_current_buffer = NULL;
  mtr->streaming = 0;

#ifdef CONFIG_V4L2_PERFORMANCE
  /* Note: should really delay this till next capture */
  mtr->perf.frames = 0;
  mtr->perf.framesdropped = 0;
  mtr->perf.bytesout = 0;
#endif // CONFIG_V4L2_PERFORMANCE
}


/**     Read out and convert the next frame
 * @return length of data or negative for error
 */
static int
capture_imagereadout (struct meteor *mtr,
		      __u8 * capture_buffer,
		      __u8 * output_buffer,
		      int output_size,
		      int output_is_user)
{
  int len;
    
  len = translate_image(mtr, capture_buffer, output_buffer,
		      output_size, output_is_user);
   
  if (len < 0)
    return len;
    
#ifdef CONFIG_V4L2_PERFORMANCE
  ++mtr->perf.frames;
  mtr->perf.bytesout += len;
#endif // CONFIG_V4L2_PERFORMANCE

  return len;
}

/*  The hardware has issued the interrupt signal, do any post-capture
 *  processing that may be necessary.
 *  [This function is called indirectly through the immediate task queue;
 *  it executes at elevated IRQL, but it is interruptible. (It's a b.h.)]
 */



static void
capture_interrupt (void *v)
{
  register struct meteor *mtr = (struct meteor *) v;
  register struct saa7116 *s7116 = (struct saa7116*)mtr->s7116;
  struct stream_buffer *buf;
  stamp_t timestamp_rough;
  __u32 ts_sec, ts_usec;
  __u32 size, len, buffer, islast;
  __u32 stream_frame_num;
#ifdef RATE_CONTROL
  __u32 raw_frame_num, next_raw_frame_to_keep, stream_frame_num;
  stamp_t temp64;
#endif

  if (!mtr->grabber_enabled || mtr->capture_completed)
    return;
  
  if ( (mtr->capture_buffer == NULL &&
        mtr->stream_current_buffer == NULL) ||
       !mtr->ints_enabled )
    {
      debug_msg ("meteor%d: Can't process the interrupt\n",mtr->index);
      return;
    }

  saa7116_check(mtr);     
  
  switch (mtr->flags & METEOR_HAVE_MASK) 
   {
    case 0:
      debug_msg("meteor%d: Spurious interrupt\n",mtr->index);
      return;
      break;
    case METEOR_HAVE_EVEN:
      mtr->flags &= ~METEOR_HAVE_EVEN;
      break;
    case METEOR_HAVE_ODD:
      mtr->flags &= ~METEOR_HAVE_ODD;
      break;
    case METEOR_HAVE_MASK:
      mtr->flags &= ~METEOR_HAVE_MASK;
      break;      
   }
  

  if (!mtr->streaming)
    {
      /* capture completed should only be used for read */
      mtr->capture_completed = 1;
      mtr->time_acquired = current_time_ms ();
      /* DMA might not have finished, but we'll check in read() */
#ifdef METEOR_NEW_WAIT 
      if (waitqueue_active(&mtr->new_video_frame)) 
        wake_up(&mtr->new_video_frame);
#else
      wake_up_interruptible (&mtr->new_video_frame);
#endif
      return;
    }
  else
   {
    /*  Only get here in streaming mode  */
    if (mtr->stream_last_frame == -2)
      {	/*     First frame of the stream  */
        /* set the timestamp of the beginning of the stream [units 1ns] */
        v4l2_masterclock_gettime (&mtr->stream_begin);
        mtr->stream_last_frame = -1;
      }

    #ifdef METEOR_USE_LIST
      islast = meteor_list_is_last( &mtr->stream_q_capture );
    #else
      islast = v4l2_q_last(&mtr->stream_q_capture);
    #endif
    if (islast == 1)
      {
         /* No available buffers. Skip this frame. This is not an  */
         /* error, it's a normal way to throttle the capture rate  */
         #ifdef CONFIG_V4L2_PERFORMANCE
	   mtr->perf.framesdropped++;
         #endif // CONFIG_V4L2_PERFORMANCE
         return;
      }
    else if (islast == 0)
      {
        buf = mtr->stream_current_buffer;
      }
    else 
      {
	 return;
      }

    /*  Compute current stream time  [units 1ns] */
    v4l2_masterclock_gettime (&timestamp_rough);
    /*  We do computations on a relative timestamp (since beginning  */
    /*  of stream), then add 'mtr->stream_begin' back in later       */
    timestamp_rough -= mtr->stream_begin;

#ifdef RATE_CONTROL
    /*  Capture rate control: do not work */
    if (mtr->videc.frame_period.numerator == 0 ||
	mtr->videc.frame_period.denominator == 0 )
      {	//     Sanity check to avoid divide-by-zero 
#ifdef DEBUG
        debug_msg ("meteor%d: videc.frame_period == %d %d -- divide by 0\n",
                 mtr->index, 
		 mtr->videc.frame_period.numerator, 
		 mtr->videc.frame_period.denominator );
#endif
        mtr->videc.frame_period = {1001, 30000}; // 333667;
        mtr->frame_period = v4l2_fract_to_100ns( mtr->videc.frame_period );
      }
    // NOTE timestamp is in units ns
    //      timeperframe is in units 100ns
    //      frame_period is in units 100ns
    raw_frame_num = v4l2_timestamp_divide( timestamp_rough, mtr->frame_period );
    temp64 = (stamp_t) mtr->timeperframe
      * (mtr->stream_last_frame + 1)
      + (mtr->videc.frame_period >> 1);
    next_raw_frame_to_keep = v4l2_math_div6432 ((u64) temp64, mtr->frame_period, NULL);

    // Why I need to multiply by 2 raw_frame_num 
    if (2*raw_frame_num < next_raw_frame_to_keep)
      {	//     Not time yet, don't keep this frame 
        info_msg("Not time yet\n");
        return;
      }
#endif

    /*  Want this frame */
    len = buf->himem_node->len;
    
#ifdef CONFIG_V4L2_PERFORMANCE
    ++mtr->perf.frames;
    mtr->perf.bytesout += len;
#endif // CONFIG_V4L2_PERFORMANCE
     
    /*  Fill in the buffer information fields  */
    buf->vidbuf.bytesused = len;
    buf->vidbuf.flags |= V4L2_BUF_FLAG_DONE | V4L2_BUF_FLAG_KEYFRAME;
 
    /* timestamp (in 1ns) and timeperframe (in 100ns) */
    stream_frame_num = 
	    v4l2_timestamp_correct (&timestamp_rough, mtr->timeperframe );
    /*  Returned timestamp should be absolute system time  */
    timestamp_rough +=  mtr->stream_begin;  /* make stamp absolute */
    ts_sec = v4l2_math_div6432 ((u64)timestamp_rough, 1000000000, &ts_usec);
    ts_usec /= 1000;
    buf->vidbuf.timestamp = (struct timeval){ ts_sec, ts_usec };
    buf->vidbuf.sequence = stream_frame_num;

    #ifdef CONFIG_V4L2_PERFORMANCE
      if (stream_frame_num > mtr->stream_last_frame + 1)
      {				/*     We have missed one or more frames  */
        mtr->perf.framesdropped += stream_frame_num
  	  - (mtr->stream_last_frame + 1);
      }
    #endif // CONFIG_V4L2_PERFORMANCE
    mtr->stream_last_frame = stream_frame_num;

    /*  Move buffer to done queue  */
    #ifdef METEOR_USE_LIST
      meteor_list_move_head_to_tail
	      ( &mtr->stream_q_capture, &mtr->stream_q_done);
    #else
      buf = v4l2_q_del_head (&mtr->stream_q_capture);
      v4l2_q_add_tail (&mtr->stream_q_done, &buf->qnode);
    #endif
    
    /* Look if there is still buffer */
    #ifdef METEOR_USE_LIST
      buf = meteor_list_peek_head(&mtr->stream_q_capture);
    #else
      buf = v4l2_q_peek_head (&mtr->stream_q_capture);
    #endif

    if (buf == NULL)
    {
        /*     No available buffers. Should not arrive here cause  */
	/*     we have verified that buffer wasn't alone.          */
        debug_msg("Danger: unforeseen buffer loss\n");
	return;
    }
    
    mtr->stream_current_buffer = buf;
    buffer = buf->himem_node->addr;
    size = buf->himem_node->len;

    s7116->dma_e[0] = mtr->dma_add_e[0] + buffer;
    s7116->dma_e[1] = mtr->dma_add_e[1] + buffer;
    s7116->dma_e[2] = mtr->dma_add_e[2] + buffer;
    s7116->dma_o[0] = mtr->dma_add_o[0] + buffer;
    s7116->dma_o[1] = mtr->dma_add_o[1] + buffer;
    s7116->dma_o[2] = mtr->dma_add_o[2] + buffer;
    s7116->stride_e[0] = mtr->dma_str_e[0];
    s7116->stride_e[1] = mtr->dma_str_e[1];
    s7116->stride_e[2] = mtr->dma_str_e[2];
    s7116->stride_o[0] = mtr->dma_str_o[0];
    s7116->stride_o[1] = mtr->dma_str_o[1];
    s7116->stride_o[2] = mtr->dma_str_o[2];
    s7116->dma_end_e = s7116->dma_end_o = (buffer + size - 1);
      
    mtr->grabber_enabled = 1;

    /*  A new frame is ready!  */
#ifdef METEOR_NEW_WAIT 
    if (waitqueue_active(&mtr->new_video_frame)) 
      wake_up(&mtr->new_video_frame);
#else
    wake_up_interruptible (&mtr->new_video_frame);
#endif 
  }
}


/*  Read captured data into a user buffer.
 *  Return: negative = error
 *          0        = keep waiting
 *          positive = count of bytes read successfully
 */
static long
capture_read (struct meteor *mtr,
	      __u8 * user_buffer,
	      int user_buffer_size)
{
  int len = user_buffer_size;
  __u32 now;

  if (!mtr->ints_enabled)
    return -EIO;     
  
  if (!mtr->capture_completed)
    {/* No interrupt has occurred yet, or DMA didn't finish.  */
      debug_msg ("No data ready.\n");
      if (!mtr->grabber_enabled)
	capture_grab_frame (mtr);
      return 0;			/* caller should keep waiting */
    }

  now = current_time_ms ();
  if (now - mtr->time_acquired > MAX_FRAME_AGE)
    {/* Frame in buffer is stale, get a new one */
      debug_msg ("Stale frame, re-acquiring.\n");
      mtr->grabber_enabled = 0;
      grabbing_enable (mtr, 0);
      capture_grab_frame (mtr);
      return 0;			/* caller should keep waiting */
    }

  len = capture_imagereadout (mtr, mtr->capture_buffer,
			      user_buffer, user_buffer_size, 1);

  return len;
}

/*  Stop capturing and free all resources used for capture.
 */
static void
capture_close (struct meteor *mtr)
{
  int i;

  if (mtr->streaming) {
    capture_streamoff (mtr, V4L2_BUF_TYPE_VIDEO_CAPTURE);
    /* (mc) capture_abort is already called in capture_streamoff() */
  } else {
    capture_abort (mtr);
  }
  mtr->ready_to_capture = 0;
  translate_close(mtr);
  if (mtr->capture_buffer != NULL)
   {
    free_himem (mtr->capture_buffer, mtr->capture_himem_node);
    mtr->capture_buffer = NULL;
   }

  for (i = 0; i < MAX_NUM_FRAMES; ++i)
    {
      mtr->stream_buf[i].requested = 0;
      if (mtr->stream_buf[i].vaddress)
       {
	free_himem (mtr->stream_buf[i].vaddress, mtr->stream_buf[i].himem_node);
        mtr->stream_buf[i].vaddress = NULL;
       }
    }
}

/**  Reset all resources used for capture..
 */
static void
capture_open (struct meteor *mtr)
{
  int i;

  mtr->capture_buffer = NULL;
  mtr->capture_buffer_size = 0;

  for (i = 0; i < MAX_NUM_FRAMES; ++i)
   {
    mtr->stream_buf[i].requested = 0;
    mtr->stream_buf[i].vaddress = NULL;
   }

  mtr->fifo_errors = 0;
  mtr->dma_errors = 0;
  mtr->capture_completed = 0;
  mtr->grabber_enabled = 0;
}

/* ==============================================================
 *
 *      I N T E R R U P T   R O U T I N E S
 *
 * ============================================================== */


/*  Called by the system when our hardware interrupt occurs
 *  (mc) changed return type and added values to the returns
 */
static irqreturn_t
interrupt_hw (int irq, void *dev_id, struct pt_regs *fake)
{
  __u32 ints_status, clear = 0;
  register struct saa7116 *s7116;
  register struct meteor *mtr;  

  if (dev_id == NULL)
    return IRQ_NONE;

  mtr = (struct meteor *) dev_id;
    
  /* Check for an interrupt pending on the device, and  */
  /* launch handler if there is interrupt pending       */
  if (!mtr->ints_enabled)
    return IRQ_NONE;

  /* Check for an interrupt pending on the device, and  */
  /* launch handler if there is interrupt pending       */
  s7116 = mtr->s7116;
  ints_status = s7116->ints_ctl;
  if (!(ints_status & 0x07))
    return IRQ_NONE;
  
  /* Clear the interrupt */
  if (ints_status & 0x1) 
   {
    mtr->flags |= METEOR_HAVE_EVEN;
    clear |= 0x01; /* Clear End of Even Field Interrupt Status */
    s7116->capt_ctl |= 0x10; /* Clear Even Field Done */
   }
  if (ints_status & 0x2) 
   {
    mtr->flags |= METEOR_HAVE_ODD;
    clear |= 0x02; /* Clear End of Odd Field Interrupt Status */
    s7116->capt_ctl |= 0x20; /* Clear Even Field Done */
   }
  if (ints_status & 0x4) 
   {
    /* Should not arrive here in principle ! */
    clear |= 0x04; /* Clear Start of Field Interrupt Status */
   }
   
  /* If interlaced wait for the two fields */  
  if ( ( (mtr->clientfmt.field == V4L2_FIELD_INTERLACED ) 
         && ((mtr->flags & METEOR_HAVE_MASK) == (METEOR_HAVE_EVEN | METEOR_HAVE_ODD)) )
       || !(mtr->clientfmt.field == V4L2_FIELD_INTERLACED) )
    {

      /*  Call "bottom half" of handler
        Suppressed caused I need to change the capture buffer before
        starting capture of next stream and this is done in capture_interrupt
        TODO: Is there a better way ?
        mtr->tqnode_dpc.routine = capture_interrupt;
        mtr->tqnode_dpc.data = mtr;
        queue_task (&mtr->tqnode_dpc, &tq_immediate);
        mark_bh (IMMEDIATE_BH);
      */
      #ifdef METEOR_USE_WORKQUEUE
	PREPARE_WORK( &meteor_wq, capture_interrupt, mtr);
        // schedule_work( &meteor_wq );
	queue_work( meteor_wq_thread, &meteor_wq );
      #else
        capture_interrupt(mtr);
      #endif
    }
     
    
  s7116->ints_ctl |= clear; /* Clear End of Even Field Interrupt Status */

  return IRQ_HANDLED;
}


static void
interrupt_disable (struct meteor *mtr)
{
  register struct saa7116 *s7116 = mtr->s7116;

  if (!mtr->ints_enabled)
    return;
  mtr->ints_enabled = 0;

  /*  Disable all interrupts on the device  */
  s7116->ints_ctl = 0x07;
  
  /*  Wake up any processes that might be waiting for a frame  */
  /*  and let them return an error  */
#ifdef METEOR_NEW_WAIT
  if (waitqueue_active(&mtr->new_video_frame)) 
    wake_up(&mtr->new_video_frame);
#else
  wake_up_interruptible (&mtr->new_video_frame);
#endif
}

static void
interrupt_enable (struct meteor *mtr)
{
 __u32 intr;
  
 if (mtr->irq == 0)
   {
    debug_msg ("No IRQ: cannot enable interrupts\n");
    return;
   }
    
 if (mtr->ints_enabled)
   interrupt_disable (mtr);

 mtr->ints_enabled = 1;

 switch (mtr->clientfmt.field)
  {
   case V4L2_FIELD_BOTTOM:
     intr = 0x107;
     break;
   case V4L2_FIELD_TOP:
     intr = 0x207;
     break;
   default:
     intr = 0x307;
     break;
  }

 mtr->s7116->ints_ctl = intr;
}

/* ==============================================================
 *
 *      M E M O R Y   M A P P I N G
 *
 * ============================================================== */

static struct stream_buffer *
mmap_stream_buffer_from_offset (struct meteor *mtr, __u32 offset)
{
  int i;
 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,25)
  offset = offset * PAGE_SIZE;
#endif

  for (i = 0; i < MAX_NUM_FRAMES; ++i)
    if (offset == mtr->stream_buf[i].vidbuf.m.offset)
      return &mtr->stream_buf[i];
  return NULL;
}

static int
mmap_request_buffers (struct meteor *mtr,
		      struct v4l2_requestbuffers *req)
{
  int i;
  u32 buflen;
  u32 type;

  if (mtr->stream_buffers_mapped)
    return 0;			/* can't make requests if buffers are mapped */
  if (req->count < 1)
    req->count = 1;
  if (req->count > MAX_NUM_FRAMES)
    req->count = MAX_NUM_FRAMES;
  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  mtr->stream_contig_map = 0;
  if (req->type & V4L2_BUF_REQ_CONTIG)
    {
      mtr->stream_contig_map = 1;
      req->type = type | V4L2_BUF_REQ_CONTIG;
      /* note: _REQ_CONTIG is only used in v4l2_requestbuffers */
    }
 
  /*  The buffer length needs to be a multiple of the page size  */
  buflen = (mtr->clientfmt.sizeimage + PAGE_SIZE - 1)
    & ~(PAGE_SIZE - 1);

  // debug_msg ("Granting %d buffers\n", req->count);

  /*  Now initialize the buffer structures. Don't allocate the */
  /*  buffers until they're mapped. */
  for (i = 0; i < req->count; ++i)
    {
      mtr->stream_buf[i].requested = 1;
      mtr->stream_buf[i].vidbuf.index = i;
      mtr->stream_buf[i].vidbuf.type = type;
       /* offset must be unique for each buffer, and a multiple */
       /* of PAGE_SIZE on 2.4.x  */
      mtr->stream_buf[i].vidbuf.m.offset = PAGE_SIZE * (i+1);
      mtr->stream_buf[i].vidbuf.length = buflen;
      mtr->stream_buf[i].vidbuf.bytesused = 0;
      mtr->stream_buf[i].vidbuf.flags = 0;
      mtr->stream_buf[i].vidbuf.timestamp = (struct timeval){0, 0};
      mtr->stream_buf[i].vidbuf.sequence = 0;
      memset (&mtr->stream_buf[i].vidbuf.timecode, 0,
	      sizeof (struct v4l2_timecode));
    }
  for (i = req->count; i < MAX_NUM_FRAMES; ++i)
    mtr->stream_buf[i].requested = 0;
  mtr->stream_buffers_requested = req->count;

  return 1;
}

static void
mmap_unrequest_buffers (struct meteor *mtr)
{
  int i;

  if (mtr->stream_buffers_requested == 0 ||
      mtr->stream_buffers_mapped)
    return;
  for (i = 0; i < MAX_NUM_FRAMES; ++i)
    mtr->stream_buf[i].requested = 0;
  mtr->stream_buffers_requested = 0;
}

static void
mmap_vma_open (struct vm_area_struct *vma)
{
  struct device_open * o;
  struct meteor * mtr = meteor_from_file ( NULL, vma->vm_file, &o);
  if ( mtr == NULL )
    return;
}

static void
mmap_vma_close (struct vm_area_struct *vma)
{
  struct device_open * o;
  struct meteor * mtr = meteor_from_file ( NULL, vma->vm_file, &o);
  struct stream_buffer * buf =
  mmap_stream_buffer_from_offset (mtr, vma->vm_pgoff);
  int i, n = 1;

  if (mtr->stream_contig_map)
    {				/*     Unmap all the buffers in one stroke  */
      n = mtr->stream_buffers_mapped;
      buf = &mtr->stream_buf[0];
    }
  for (i = 0; i < n; ++i)
    {
      if (mtr->streaming)
      {
	  info_msg ("Warning: munmap() called while streaming\n");
	  capture_streamoff (mtr, buf->vidbuf.type);
      }
      #ifdef METEOR_USE_LIST
        meteor_list_try_del( &buf->qnode, &mtr->stream_q_capture );
        meteor_list_try_del( &buf->qnode, &mtr->stream_q_done );
      #else
        v4l2_q_yank_node (&mtr->stream_q_capture, &buf->qnode);
        v4l2_q_yank_node (&mtr->stream_q_done, &buf->qnode);
      #endif
        
      if (buf->vaddress != NULL && i == 0)
        {
          /* Current buffer is reused as a capture buffer */
          if (buf == mtr->stream_current_buffer) 
	    {
	      mtr->capture_buffer = buf->vaddress;
	      mtr->capture_himem_node = buf->himem_node;
	      mtr->capture_buffer_size = buf->himem_node->len;
	    }
	  else
            {
  	      free_himem (buf->vaddress, buf->himem_node);
            }
	    
        }
      buf->vaddress = NULL;
      buf->vidbuf.flags = 0;
      // debug_msg("Buffer %d deallocated\n",(int)(vma->vm_pgoff/PAGE_SIZE));
      ++buf;
      if (mtr->stream_buffers_mapped > 0)
	--mtr->stream_buffers_mapped;
    }
}

static struct vm_operations_struct capture_vma_operations = {
	.open  = mmap_vma_open, 
	.close = mmap_vma_close
		// no page()
		// no populate()
};


/* ==============================================================
 *
 *   M E T E O R   V I D E O   F O R   L I N U X   I N T E R F A C E
 *
 * ============================================================== */

static int
meteor_v4l2_close ( struct inode * inode, struct file * file )
{
  struct meteor * mtr;
  struct video_device * vd;
  struct device_open * o;
  
  mtr = meteor_from_file( inode, file, &o );
  if ( NULL == mtr ) 
	  return -ENODEV;

  vd = mtr->video_dev;
  // FIXME (mc) should release vd ?

  if (!o->noncapturing)
    {
      --mtr->capturing_opens;
      if (mtr->capturing_opens == 0)
        o->noncapturing = 1;
    }
  o->isopen = 0;
  o->dev = NULL;
  --mtr->open_count;
  if (mtr->open_count == 0)
    {
      interrupt_disable (mtr);
      capture_close (mtr);
      if (mtr->irq)
	free_irq (mtr->irq, mtr);
      mtr->irq = 0;
    }
  file->private_data = NULL;

  return 0;
}

static int
meteor_v4l2_open ( struct inode * inode, struct file * file )
{
  int i, n;
  int cap;
  int flags = file->f_flags;
  struct meteor * mtr;
  struct video_device * vd;
#ifdef DEBUG
  int minor = iminor(inode);
#endif

  vd  = video_devdata( file );
  mtr = video_get_drvdata( vd );

  for (i = 0, n = -1, cap = 0; i < MAX_OPENS; ++i)
    {
      if (!mtr->open_data[i].isopen)
	n = i;			/* available open_data structure */
      else if (!mtr->open_data[i].noncapturing)
	cap = 1;		/* another open is already capturing */
    }
  if (n == -1)			/* No available open_data structures */
    {
      debug_msg ("No more opens on this device. Minor %d\n", minor);
      return -EBUSY;
    }

  if (flags & O_TRUNC)		/*  Non-capturing open */
      mtr->open_data[n].noncapturing = 1;
  else if (cap)
    {
      debug_msg ("No more capturing opens on this device\n");
      return -EBUSY;
    }
  else
    {
      mtr->open_data[n].noncapturing = 0;
      /*  Keep track of whether there is a capturing open  */
      ++mtr->capturing_opens;
#ifdef CONFIG_V4L2_PERFORMANCE
      mtr->perf.frames = 0;
      mtr->perf.framesdropped = 0;
      mtr->perf.bytesout = 0;
#endif // CONFIG_V4L2_PERFORMANCE
    }

  ++mtr->open_count;  // meteor open count
  mtr->open_data[n].isopen = 1;
  mtr->open_data[n].dev = mtr;
  file->private_data = &mtr->open_data[n];

  if (mtr->open_count == 1)
    {
       if (mtr->pci && mtr->irq == 0) 
	{
	  mtr->irq = mtr->pci->irq;
	  if ( request_irq (mtr->irq, interrupt_hw, SA_INTERRUPT | SA_SHIRQ,
		   mtr->shortname, mtr) < 0)
	    {
	      debug_msg ("Denied IRQ %d\n", mtr->irq);
              return -ENOSYS;
	    }
	}
      capture_open(mtr);
      #ifdef METEOR_USE_LIST
        INIT_LIST_HEAD( &mtr->stream_q_capture );
	INIT_LIST_HEAD( &mtr->stream_q_done );
      #else
        v4l2_q_init (&mtr->stream_q_capture);
        v4l2_q_init (&mtr->stream_q_done);
      #endif
    }      

  /* frame counter for test images only */
  if (!mtr->open_data[n].noncapturing)
    mtr->h = mtr->m = mtr->s = mtr->f = 0;
  return 0;
}




static ssize_t
meteor_v4l2_write ( struct file * file, const char *buf, size_t count, loff_t * ppos)
{
  debug_msg ("Write() not handled\n");
  return -EINVAL;
}

/* The arguments are already copied into kernel memory, so don't use
   copy_from_user() or copy_to_user() on arg.  
   FIXME (mc) check ioctl
 */
static int
meteor_v4l2_ioctl ( struct inode * inode, struct file * file, 
		unsigned int cmd, unsigned long arg)
{
  struct meteor * mtr;
  struct device_open * o;
  void * parg = (void *)arg;

  debug_ioctl ("ioctl cmd: %d arg: %x\n", _IOC_NR (cmd), (void *)arg );

  mtr = meteor_from_file( inode, file, &o );
  if ( NULL == mtr ) 
	return -EINVAL;

  switch (cmd)
    {
    case VIDIOC_QUERYCAP:
      {
        debug_ioctl ("ioctl VIDIOC_QUERYCAP \n");
	struct v4l2_capability * b = parg;
	strcpy (b->driver, mtr->video_dev->name);
	b->capabilities = V4L2_CAP_VIDEO_CAPTURE
		        | V4L2_CAP_READWRITE
			| V4L2_CAP_STREAMING;
	b->version = KERNEL_VERSION(2,6,2);
	strcpy (b->card, "Meteor" );  // FIXME (mc) do better
	strcpy (b->bus_info, "PCI:"); // FIXME (mc) do better

	/*
	strcpy (b->name, mtr->v.name);
	b->type = V4L2_TYPE_VIDEO_CAPTURE;
	b->flags = V4L2_FLAG_READ | V4L2_FLAG_STREAMING | V4L2_FLAG_SELECT;
	b->inputs = mtr->videc.num_inputs;
	b->outputs = 0;
	b->audios = 0;
	b->maxwidth = MAX_WIDTH;
	b->maxheight = MAX_HEIGHT;
	b->minwidth = MIN_WIDTH;
	b->minheight = MIN_HEIGHT;
	b->maxframerate = 30;
	*/
	return 0;
      }

    /* (mc) dropped 
    case VIDIOC_ENUM_PIXFMT:
      {
	struct v4l2_fmtdesc *f = parg;
	if (f->index < 0 || f->index >= NUM_CAPFMT)
	  return -EINVAL;
	*f = capfmt[f->index];
	return 0;
      }
    */

    case VIDIOC_G_FMT:
      {
	struct v4l2_format *fmt = parg;
        debug_ioctl ("ioctl VIDIOC_G_FMT \n");
	if (fmt->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
	  {
	    debug_msg ("G_FMT wrong buffer type %d\n", fmt->type);
	    return -EINVAL;
	  }
	fmt->fmt.pix = mtr->clientfmt;
	return 0;
      }

    case VIDIOC_S_FMT:
      {
	struct v4l2_format *fmt = parg;
        debug_ioctl ("ioctl VIDIOC_S_FMT \n");
	if (o->noncapturing)
	  {
	    debug_msg ("S_FMT illegal in non-capturing open\n");
	    return -EPERM;
	  }
	if (fmt->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
	  {
	    debug_msg ("S_FMT wrong buffer type %d\n", fmt->type);
	    return -EINVAL;
	  }
	mtr->clientfmt = fmt->fmt.pix;

        /* TODO: Needed ? */
	if (mtr->streaming)
          capture_abort(mtr);
	  
	if (!capture_new_format (mtr))
	  return -EINVAL;
	if (mtr->streaming)
	  capture_grab_frame (mtr);
	fmt->fmt.pix = mtr->clientfmt;
	return 0;
      }

    /* (mc) dropped see videodev2.h
    case VIDIOC_G_COMP:  // 6
      return -EINVAL;
    case VIDIOC_S_COMP:  // 7
      return -EINVAL;
    */

    case VIDIOC_REQBUFS:
      {
	struct v4l2_requestbuffers *req = parg;
        debug_ioctl ("ioctl VIDIOC_REQBUFS \n");
	if (o->noncapturing)
	  {
	    debug_msg ("REQBUFS illegal in non-capturing open\n");
	    return -EPERM;
	  }
	if (mtr->stream_buffers_mapped)
	  {
	    debug_msg ("Can't request buffers if buffers are "
	 	       "already mapped\n");
	    return -EPERM;
	  }
	mmap_unrequest_buffers (mtr);
	capture_begin (mtr);
	if (!mmap_request_buffers (mtr, req))
	  return -EINVAL;
	return 0;
      }

    case VIDIOC_QUERYBUF:
      {
	struct v4l2_buffer *buf = parg;
	int i;
        debug_ioctl ("ioctl VIDIOC_QUERYBUF \n");
	if (o->noncapturing)
	  {
	    debug_msg ("QUERYBUF illegal in non-capturing open\n");
	    return -EPERM;
	  }
	i = buf->index;
	if (i < 0 || i >= MAX_NUM_FRAMES ||
	    !mtr->stream_buf[i].requested ||
	    (buf->type & V4L2_BUF_TYPE_field) !=
	    (mtr->stream_buf[i].vidbuf.type & V4L2_BUF_TYPE_field))
	  {
	    debug_msg ("QUERYBUF bad parameter\n");
	    return -EINVAL;
	  }
	*buf = mtr->stream_buf[i].vidbuf;
	return 0;
      }

    case VIDIOC_QBUF:
      {
	struct v4l2_buffer *buf = parg;
	debug_ioctl ("ioctl VIDIOC_QBUF \n");
	if (o->noncapturing)
	  {
	    debug_msg ("QBUF illegal in non-capturing open\n");
	    return -EPERM;
	  }
	if (!mtr->stream_buffers_mapped)
	  {
	    debug_msg ("QBUF no buffers are mapped\n");
	    return -EINVAL;
	  }
	if (!capture_queuebuffer (mtr, buf))
	  return -EINVAL;
	return 0;
      }

    case VIDIOC_DQBUF:
      {
	struct v4l2_buffer *buf = parg;
	debug_ioctl ("ioctl VIDIOC_DQBUF \n");
	if (o->noncapturing)
	  {
	    debug_msg ("DQBUF illegal in non-capturing open\n");
	    return -EPERM;
	  }
	if (!capture_dequeuebuffer (mtr, buf))
	  return -EINVAL;
	return 0;
      }

    case VIDIOC_STREAMON:
      {
	__u32 *type = parg;
	
	debug_ioctl ("ioctl VIDIOC_STREAMON \n");
	if (o->noncapturing)
	  {
	    debug_msg ("STREAMON illegal in non-capturing open\n");
	    return -EPERM;
	  }
	  
	if (!capture_streamon (mtr, *type))
	  return -EINVAL;
	return 0;
      }

    case VIDIOC_STREAMOFF:
      {
	__u32 *type = parg;
	debug_ioctl ("ioctl VIDIOC_STREAMOFF \n");
	if (o->noncapturing)
	  {
	    debug_msg ("STREAMOFF illegal in non-capturing open\n");
	    return -EPERM;
	  }
	capture_streamoff (mtr, *type);
	return 0;
      }

    /* (mc) dropped
    case VIDIOC_ENUM_FBUFFMT:  // 3
      return -EINVAL;
    */
    case VIDIOC_G_FBUF:
      debug_ioctl ("ioctl VIDIOC_G_FBUF\n");
      return -EINVAL;
    case VIDIOC_S_FBUF:
      debug_ioctl ("ioctl VIDIOC_S_FBUF\n");
      return -EINVAL;
#if 0 //  (mc)
    case VIDIOC_G_WIN:  // 12
      {
        struct v4l2_window	*win = parg;

        win->x =  (SAA7196_REG(mtr, SAA7196_HWS) & 0xff) /* ?? - 2*/ + 
                 ((SAA7196_REG(mtr, SAA7196_HFILT) & 0x10)<<8);
	win->y = (SAA7196_REG(mtr, SAA7196_VWS) & 0xff) /* ?? - 14 */ +
                 ((SAA7196_REG(mtr, SAA7196_VYP) & 0x10)<<8);
        if ((mtr->clientfmt.flags & V4L2_FMT_FLAG_FIELD_field) == V4L2_FIELD_INTERLACED)
	  win->y = (win->y)>> 1;
	  
	win->width = mtr->clientfmt.width;
	win->height = mtr->clientfmt.height;
	win->chromakey = 0;
	win->clips = NULL;
	win->clipcount = 0;
	return 0;
      }
    case VIDIOC_S_WIN:  // 13
      {
        struct v4l2_window	*win = parg;

        if ( ((win->x > 0) && (win->x < mtr->clientfmt.width)) &&
	     ((win->y > 0) && (win->y < mtr->clientfmt.height)) )
	 {
	  int lines;
	  
          SAA7196_WRITE(mtr, SAA7196_HWS,   (win->x & 0xff));
          SAA7196_WRITE(mtr, SAA7196_HFILT, (SAA7196_REG(mtr, SAA7196_HFILT) & ~0x10) |
	                                    (win->x & 0x100)>>4);
					    
          /* Vertical windows start in line number per field */
          if ((mtr->clientfmt.flags & V4L2_FMT_FLAG_FIELD_field) == V4L2_FIELD_INTERLACED)
	    lines = (win->y)>>1;
	  else
	    lines = win->y;
	  
          SAA7196_WRITE(mtr, SAA7196_VWS,   (lines & 0xff));
          SAA7196_WRITE(mtr, SAA7196_VYP,   (SAA7196_REG(mtr, SAA7196_VYP) & ~0x10) |
	                                    (lines & 0x100)>>4);
	 }
	else
	  return -EINVAL;
	
	if (win->clips != NULL)
	  return -EINVAL;
	  
	return 0;
      }
#endif
#if 0 // (mc)
    case VIDIOC_PREVIEW:  // 14
      return -EINVAL;
#else
    case VIDIOC_OVERLAY:
      debug_ioctl ("ioctl VIDIOC_OVERLAY\n");
      return -EINVAL;
#endif

#ifdef CONFIG_V4L2_PERFORMANCE
    case VIDIOC_G_PERF: // 20
      debug_ioctl ("ioctl VIDIOC_G_PERF\n");
      memcpy (parg, &mtr->perf, sizeof (mtr->perf));
      return 0;
#endif // CONFIG_V4L2_PERFORMANCE

    case VIDIOC_G_INPUT:
      debug_ioctl ("ioctl VIDIOC_G_INPUT\n");
      memcpy (parg, &mtr->input, sizeof (mtr->input));
      return 0;

    case VIDIOC_S_INPUT:
      {
	int input = *(int *) parg;
	debug_ioctl ("ioctl VIDIOC_S_INPUT\n");
	if (input < 0 || input >= mtr->videc.num_inputs)
	  {
	    debug_msg ("Input out of range %d\n", input);
	    return -EINVAL;
	  }
	if (input != mtr->input)
	  {
	    mtr->input = input;
            mtr->videc.set_input (mtr, input);
	  }
	return 0;
      }

    case VIDIOC_G_PARM:
      {
	struct v4l2_streamparm *sp = parg;
	debug_ioctl ("ioctl VIDIOC_G_PARM\n");
	if (sp->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
	  return -EINVAL;
	sp->parm.capture = mtr->capture;
	return 0;
      }

    case VIDIOC_S_PARM:
      {
        /* TODO: Field mask control according to wanted fps */
        /* TODO: Automatic Gain Control allows three speeds slow/medium/fast */
	/* TODO: Horizontal decimation filter control */
        /* TODO: Vertical Luminance data processing */
        /* TODO: Inverse luminance and 2's complement U V */
        //  SAA7196_WRITE (mtr, SAA7196_DPATH, (SAA7196_REG (mtr, SAA7196_DPATH) & ~0x10));
        //  SAA7196_WRITE (mtr, SAA7196_DPATH, (SAA7196_REG (mtr, SAA7196_DPATH) | 0x10));
        /* TODO: Gamma Bypass bit control here set false by default */
        //  SAA7196_WRITE (mtr, SAA7196_FMTS, (SAA7196_REG (mtr, SAA7196_FMTS) & ~0x80));
        //  SAA7196_WRITE (mtr, SAA7196_FMTS, (SAA7196_REG (mtr, SAA7196_FMTS) | 0x80));

	struct v4l2_streamparm *sp = parg;
	struct v4l2_captureparm *vp = &sp->parm.capture;

	debug_ioctl ("ioctl VIDIOC_S_PARM\n");
	if (vp->capturemode & ~mtr->capture.capability)
	  {
	    debug_msg ("PARM unsupported capture capability %08X\n",
	 	       vp->capturemode);
	    return -EINVAL;
	  }
	if ((mtr->capture.capability & V4L2_CAP_TIMEPERFRAME) &&
	    (vp->timeperframe.numerator * 1000 < vp->timeperframe.denominator ) )
	    // N/D < 1 msec = 10000 0.1 usec
	  {
	    debug_msg ("PARM time per frame out of range: %d/%d\n",
		       vp->timeperframe.numerator, vp->timeperframe.denominator);
	    return -EINVAL;
	  }
	if (vp->capturemode != mtr->capture.capturemode &&
	    !o->noncapturing && mtr->streaming)
	  return -EINVAL;

	if (o->noncapturing)
	  return 0;
	if (vp->capturemode != mtr->capture.capturemode)
	  {
	    mtr->capture.capturemode = vp->capturemode;
	    capture_new_format (mtr);
	  }
	
        /*  Look for video input and set frame period accordingly */
        mtr->videc.is_stable(mtr);
	if ((vp->capturemode & V4L2_CAP_TIMEPERFRAME) &&
	    (! v4l2_fract_isless( vp->timeperframe, mtr->videc.frame_period )) )
	  mtr->capture.timeperframe = vp->timeperframe;
	else
	  mtr->capture.timeperframe = mtr->videc.frame_period;

	/* set mtr->timeperframe accordingly */
  	mtr->timeperframe = v4l2_fract_to_100ns( mtr->capture.timeperframe );

	return 0;
      }

    case VIDIOC_G_STD:
      {
	struct v4l2_standard *std = parg;
	debug_ioctl ("ioctl VIDIOC_G_STD\n");
	v4l2_video_std_construct (std, mtr->videc.standard, mtr->standard_name);
	return 0;
      }

    case VIDIOC_S_STD:
      {
	struct v4l2_standard *std = parg;
	debug_ioctl ("ioctl VIDIOC_S_STD\n");
	if ((o->noncapturing && mtr->capturing_opens) ||
	    mtr->stream_buffers_mapped)
	  return -EPERM;
	if ( v4l2_video_std_confirm (std) < 0) 
		return -EINVAL;
	if (!(std->id & mtr->videc.standards))
	  {
	    debug_msg ("Bad standard: %u\n", (unsigned) std->id );
	    return -EINVAL;
	  }
	if (!mtr->videc.set_standard (mtr, std->id))
	  return -EINVAL;
	return 0;
      }

    case VIDIOC_ENUMSTD:
      {
	struct v4l2_enumstd *estd = parg;
	__u32 b, i;
	debug_ioctl ("ioctl VIDIOC_ENUMSTD index %d \n", estd->index );
	if (estd->index < 0 || estd->index > 30)
	  return -EINVAL;
	for (b = 1, i = 0; b < 32; ++b)
	  {
	    int id = 1 << b;
	    if ((id & mtr->videc.standards) == 0)
	      continue;
	    if (i == estd->index)
	      {
		v4l2_video_std_construct (&estd->std, id, mtr->standard_name);
		estd->inputs = (__u32) - 1;	/* all inputs */
		estd->outputs = 0;
		return 0;
	      }
	    ++i;
	  }
	return -EINVAL;
      }

    case VIDIOC_ENUMINPUT:
      {
	struct v4l2_input *vi = parg;
	debug_ioctl ("ioctl VIDIOC_ENUMINPUT\n");
	if (vi->index < 0 || vi->index >= mtr->videc.num_inputs)
	  return -EINVAL;
	*vi = mtr->source[vi->index].input;
	return 0;
      }

    case VIDIOC_QUERYCTRL:
      {
	struct v4l2_queryctrl *qc = parg;
	int i;
	debug_ioctl ("ioctl VIDIOC_QUERYCTRL\n");
	i = find_vctrl (qc->id);
	if (i == -EINVAL)
	  {
	    qc->flags = V4L2_CTRL_FLAG_DISABLED;
	    return 0;
	  }
	if (i < 0)
	  {
	    return -EINVAL;
	  }
	/*  V4L2 filled in category and group, preserve them */
	// capture_control[i].category = qc->category;
	// memcpy (capture_control[i].group, qc->group, sizeof (qc->group));
	*qc = capture_control[i];
	return 0;
      }

    case VIDIOC_QUERYMENU:
      {
	struct v4l2_querymenu *qm = parg;

	debug_ioctl ("ioctl VIDIOC_QUERYMENU\n");
	return vctrl_querymenu (qm);
      }

    case VIDIOC_G_CTRL:
      {
	struct v4l2_control *vc = parg;
	int i;
	debug_ioctl ("ioctl VIDIOC_G_CTRL\n");
	i = find_vctrl (vc->id);
	if (i < 0)
	  return -EINVAL;
	vc->value = mtr->source[mtr->input].control[i];
	return 0;
      }

    case VIDIOC_S_CTRL:
      {
	struct v4l2_control *vc = parg;
	int i;
	debug_ioctl ("ioctl VIDIOC_S_CTRL\n");
	i = find_vctrl (vc->id);
	if (i < 0)
	  return -EINVAL;
	mtr->source[mtr->input].control[i] = vc->value;
	meteor_tone_controls (mtr);
	return 0;
      }

    case VIDIOC_G_TUNER:
      debug_ioctl ("ioctl VIDIOC_G_TUNER\n");
      return -EINVAL;
    case VIDIOC_S_TUNER:
      debug_ioctl ("ioctl VIDIOC_S_TUNER\n");
      return -EINVAL;

    case VIDIOC_G_FREQUENCY:
      debug_ioctl ("ioctl VIDIOC_G_FREQUENCY\n");
      return -EINVAL;
    case VIDIOC_S_FREQUENCY:
      debug_ioctl ("ioctl VIDIOC_S_FREQUENCY\n");
      return -EINVAL;

    case VIDIOC_G_AUDIO:
      debug_ioctl ("ioctl VIDIOC_G_AUDIO\n");
      return -EINVAL;
    case VIDIOC_S_AUDIO:
      debug_ioctl ("ioctl VIDIOC_S_AUDIO\n");
      return -EINVAL;

    default:
      debug_ioctl ("ioctl UNKNOWN CMD %d\n", _IOC_NR (cmd));
      return -ENOIOCTLCMD;
    }
  return 0;
}

static int
meteor_v4l2_mmap (struct file * file, struct vm_area_struct * vma)
{
  int i, n = 1;
  struct stream_buffer * buf;
  struct device_open * o;
  struct meteor * mtr;

  mtr = meteor_from_file( file->f_dentry->d_inode, file, &o );
  if ( NULL == mtr ) 
	return -EINVAL;

  if (o->noncapturing)
    {
      debug_msg ("mmap() called on non-capturing open\n");
      return -ENODEV;
    }
  buf = mmap_stream_buffer_from_offset (mtr, vma->vm_pgoff);
  if (mtr->stream_contig_map)
    {				/*     N buffers in one contiguous map  */
      buf = &mtr->stream_buf[0];
      n = mtr->stream_buffers_requested;
    }
  if (buf == NULL)
    {
      debug_msg ("mmap() Invalid offset parameter\n");
      return -EINVAL;		/* no such buffer */
    }
  if (buf->vidbuf.length * n != vma->vm_end - vma->vm_start)
    {
      debug_msg ("mmap() Wrong length parameter %x*%d!=%lx\n", buf->vidbuf.length, n, vma->vm_end - vma->vm_start);
      return -EINVAL;		/* wrong length */
    }

  /* All buffers are directly accessed by DMA => no
     use to reserve a stream capture buffer to be copied
     next into a buffer if in streaming mode */
  if (mtr->capture_buffer != NULL)
   {
     free_himem(mtr->capture_buffer, mtr->capture_himem_node);
     mtr->capture_buffer = NULL;
   }

  for (i = 0; i < n; ++i)
    {
      if (!buf->requested)
	{
	  debug_msg ("mmap() Buffer is not available for mapping\n");
	  return -EINVAL;	/* not requested */
	}
      if (buf->vidbuf.flags & V4L2_BUF_FLAG_MAPPED)
	{
	  debug_msg ("mmap() Buffer is already mapped\n");
	  return -EINVAL;	/* already mapped */
	}

      if (buf->vaddress != NULL)
	free_himem (buf->vaddress, buf->himem_node);
      if (i == 0)
        {
	  buf->vaddress = malloc_himem (mtr, buf->vidbuf.length * n, &(buf->himem_node));
          if (NULL == buf->vaddress)
           {
             err_msg ("meteor%d: Could not allocate mmap() buffer\n",
                      mtr->index);
             return -ENODEV;
           }

	  // (mc) added first param "vma"
	  if (remap_page_range(vma, 
			       vma->vm_start, buf->himem_node->addr, 
	                       vma->vm_end - vma->vm_start, vma->vm_page_prot))
            return -EINVAL;
        }
      else
	buf->vaddress = buf[-1].vaddress + buf->vidbuf.length;
      if (buf->vaddress == NULL)
	{
	  err_msg ("meteor%d: Could not allocate mmap() buffer\n",
                   mtr->index);
	  return -ENODEV;
        }
      buf->vidbuf.flags |= V4L2_BUF_FLAG_MAPPED;
      ++mtr->stream_buffers_mapped;
      ++buf;
    }

  vma->vm_ops = &capture_vma_operations;
  if (vma->vm_ops->open)
    vma->vm_ops->open (vma);
  /*  Note: vma->vm_file will be set up by V4L2  */

  return 0;
}

static unsigned int
meteor_v4l2_poll ( struct file * file, poll_table * table)
{
  struct device_open * o;
  struct meteor * mtr;

  mtr = meteor_from_file( file->f_dentry->d_inode, file, &o );
  if ( NULL == mtr ) 
	return -EINVAL;

  if (o->noncapturing)
    {
      debug_msg ("poll() illegal in non-capturing open\n");
      return POLLERR;
    }

  if (mtr->streaming)
    {
      void * node;
 
      #ifdef METEOR_USE_LIST
        node = meteor_list_peek_head( &mtr->stream_q_done );
      #else
        node = v4l2_q_peek_head (&mtr->stream_q_done);
      #endif
      if (node != NULL)
	return (POLLIN | POLLRDNORM);	/* data is ready now */
      #ifdef METEOR_USE_LIST
        node = meteor_list_peek_head( &mtr->stream_q_capture );
      #else
        node = v4l2_q_peek_head (&mtr->stream_q_capture);
      #endif
      if (node == NULL)
	return POLLERR;		/* no frames queued */
      poll_wait (file, &mtr->new_video_frame, table);
      saa7116_check(mtr);     
      return 0;
    }

  /*  Capture is through read() call */

  if (mtr->capture_completed)	/* data is ready now */
    return (POLLIN | POLLRDNORM);
  capture_grab_frame (mtr);	/* does nothing if capture is in progress */
  if (!mtr->ready_to_capture)	/* Can't grab frames! */
    return POLLERR;
  poll_wait (file, &mtr->new_video_frame, table);
  saa7116_check(mtr);     
  return 0;
}

static ssize_t
meteor_v4l2_read ( struct file * file, char * buf, size_t count, loff_t * pos)
{
  struct device_open * o;
  struct meteor * mtr;
  long len = 0;
  long my_timeout;

  mtr = meteor_from_file( file->f_dentry->d_inode, file, &o );
  if ( NULL == mtr ) 
	return -EINVAL;

  if (o->noncapturing)
    {
      debug_msg ("read() illegal in non-capturing open\n");
      return -EPERM;
    }
  if (mtr->streaming)
    {
      debug_msg ("Can't read() when streaming is on\n");
      return -EPERM;
    }
    
  if (!mtr->videc.is_stable(mtr))
     return -EIO;

  capture_grab_frame (mtr);	/* does nothing if capture is in progress */
  if (!mtr->ready_to_capture)
    {
      debug_msg ("Can't grab frames!\n");
      return 0;
    }

  /* Timeout sets to 200 ms */
  my_timeout = HZ / 5;
  while (len == 0)
  {
      if ( file->f_flags & O_NONBLOCK )
	{
	  if (!mtr->capture_completed)
	    return -EAGAIN;
	}
      else
	{
	  /* watch out for race condition going to sleep! */
#ifdef METEOR_NEW_WAIT 
	  DEFINE_WAIT( wait );
	  while ( !mtr->capture_completed ) 
	  {
	    prepare_to_wait( &mtr->new_video_frame, &wait, TASK_INTERRUPTIBLE);
	    if (!mtr->capture_completed )
	      schedule();
            finish_wait( &mtr->new_video_frame, &wait );
	  }

#else
	  cli ();
	  if (!mtr->capture_completed)
	    {
	      my_timeout = interruptible_sleep_on_timeout (
			      &mtr->new_video_frame, my_timeout);
	    }
	  sti ();
#endif
	}

      if (my_timeout == 0)
      {
	  err_msg ("meteor%d: Timeout on read\n", mtr->index);
	  saa7116_check(mtr);
	  break;
      }
      len = capture_read (mtr, buf, count);
  }
  return len;
}


static loff_t
meteor_v4l2_llseek ( struct file * file, loff_t offset, int origin )
{
	return -ESPIPE;
}


struct file_operations meteor_v4l2_fops = {
	.llseek = meteor_v4l2_llseek,
	.read   = meteor_v4l2_read,
	.write  = meteor_v4l2_write,
	.poll   = meteor_v4l2_poll,
	.ioctl  = meteor_v4l2_ioctl,
	.mmap   = meteor_v4l2_mmap,
	.open   = meteor_v4l2_open,
	.release =  meteor_v4l2_close,
};

/* --------------------------------------------------------------------------
 *    Remaining initialization of video decoder etc. This is only
 *    done when the device is successfully identified and registered.
 */
static void
meteor_init_done (struct meteor * mtr)
{
  int i;

  /*  Initialize video input array        */
  for (i = 0; i < VSOURCE_COUNT; ++i)
    {
      mtr->source[i].input.index = i;
      mtr->source[i].input.type = V4L2_INPUT_TYPE_CAMERA;
      // FIXME (mc)
      // mtr->source[i].input.capability = 0;
      /*  Initialize video control properties */
      mtr->source[i].control[VCTRL_BRIGHTNESS] =
	capture_control[VCTRL_BRIGHTNESS].default_value;
      mtr->source[i].control[VCTRL_CONTRAST] =
	capture_control[VCTRL_CONTRAST].default_value;
      mtr->source[i].control[VCTRL_SATURATION] =
	capture_control[VCTRL_SATURATION].default_value;
      mtr->source[i].control[VCTRL_HUE] =
	capture_control[VCTRL_HUE].default_value;
      mtr->source[i].control[VCTRL_GAIN] =
	capture_control[VCTRL_GAIN].default_value;

    }
  strcpy (mtr->source[METEOR_INPUT_DEV0].input.name, "Composite 1");
  strcpy (mtr->source[METEOR_INPUT_DEV1].input.name, "Composite 2");
  strcpy (mtr->source[METEOR_INPUT_DEV2].input.name, "Composite 3");
  strcpy (mtr->source[METEOR_INPUT_DEV3].input.name, "Composite 4");
  strcpy (mtr->source[METEOR_INPUT_DEV_SVIDEO].input.name, "S-Video");
  if (mtr->flags & METEOR_RGB)
    {
      strcpy (mtr->source[METEOR_INPUT_DEV_RGB].input.name, "RGB");
    }

  /*  Initialize the video decoder hardware       */
  mtr->videc.initialize (mtr); 
  mtr->videc.init_reg(mtr);

  if (input < 0 || input >= mtr->videc.num_inputs)
    input = 0;
  mtr->videc.set_input (mtr, input);
  mtr->videc.set_standard (mtr, standard);
  mtr->videc.set_vcrmode (mtr, 0);
  mtr->videc.is_stable(mtr);
  saa7116_fps (mtr);

  /*  Capture parameters  */
  mtr->capture.capability = V4L2_CAP_TIMEPERFRAME;
  mtr->capture.capturemode = 0;
  mtr->capture.extendedmode = 0;
  mtr->capture.timeperframe = mtr->videc.frame_period;

  mtr->frame_period = v4l2_fract_to_100ns( mtr->videc.frame_period );
  mtr->timeperframe = mtr->frame_period;

  /*  Default capture dimensions  */
  mtr->clientfmt.width = 640;
  mtr->clientfmt.height = 480;
  // FIXME (mc) no more depth field
  // mtr->depth = 32;
  mtr->clientfmt.pixelformat = V4L2_PIX_FMT_BGR32;
  mtr->clientfmt.field = V4L2_FIELD_INTERLACED;
  mtr->clientfmt.bytesperline = 0;
  mtr->clientfmt.sizeimage = 0;

  meteor_tone_controls (mtr);

  capture_new_format (mtr);
  
#ifdef METEOR_NEW_WAIT 
  init_waitqueue_head (&mtr->new_video_frame);
#else
  init_waitqueue_head (&mtr->new_video_frame);
#endif
}


/* ====================================================================
 * 
 *    D E V I C E   I N I A L I Z A T I O N   R O U T I N E S
 *
 *      These routines locate and enable the hardware, and initialize
 *      the device structure. 
 *      
 *      The functions below this point are only called during loading
 *      and unloading of the driver.
 *
 *===================================================================== */


static void
unconfig_meteor (struct meteor * mtr )
{
  if (mtr->s7116)
    {
      saa7116_reset (mtr);
      iounmap ((void *) mtr->s7116);	/* Unwind ioremap() */
    }

  // if (mtr->is_registered)
  //   {
  //     v4l2_unregister_device ((struct v4l2_device *) mtr);
  //     info_msg ("Removed device %s\n", mtr->shortname);
  //   }

  memset (mtr, 0, sizeof (meteors[0]));
}

static int
config_meteor (struct meteor *mtr, int idev)
{
  static struct pci_dev *lastpci = NULL;

  sprintf (mtr->shortname, "meteor%d", mtr - meteors);

  // (mc) removed. Deprecated
  // if (!pci_present ())
  //   return 0;

  mtr->pci = pci_find_device (PCI_VENDOR_ID_INTEL,
			      PCI_DEVICE_ID_INTEL_7116, lastpci);
  if (mtr->pci == NULL)
    return 0;			/*  No more Meteors  */
  lastpci = mtr->pci;

  mtr->index = idev;
  meteor_initialize (mtr);

  if (!p7196_probe (mtr))
    {
      err_msg ("meteor%d: Bad or unrecognized video decoder\n", mtr->index);
      unconfig_meteor ( mtr );
      return 0;			/* failed */
    }

  info_msg ("Meteor v4l2 driver : device %d irq %d\n", idev, mtr->pci->irq);

  return 1;
}


/* ====================================================================
 * 
 *    M O D U L E   I N I T   A N D   C L E A N U P
 *
 * ==================================================================== */
 

static void
handle_chipset(void)
{
	struct pci_dev *dev = NULL;
  
	if ((dev = pci_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82441, dev))) 
	{
		info_msg("Beware Meteor/PPB is required with your Chipset.\n");
	}
	else if ((dev = pci_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82437, dev))) 
	{
		info_msg("Beware Meteor (non PPB) is required with your Chispet.\n");
	}

	else if ( (dev = pci_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82443LX_0, dev)) ||
	          (dev = pci_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82443LX_1, dev)) ||
		  (dev = pci_find_device(PCI_VENDOR_ID_SI, PCI_DEVICE_ID_SI_496, dev))                 )
	{
		info_msg("Beware either Meteor/PPB or Meteor don't always work with your chipset.\n");
	}
}


// (mc) video_device' release() method 
void
meteor_v4l2_release( struct video_device * vd )
{
        debug_msg("meteor_v4l2_release video device %p\n", vd);
	if ( vd == NULL )
		return;

	struct meteor * meteor = (struct meteor *)video_get_drvdata(vd);
	if ( meteor ) {
		unconfig_meteor( meteor );
	}
	// video_device_release( vd ); // is this needed ?
	// FIXME need to free meteor ???
}

static int __init
meteor_v4l2_init (void)
{
  int i;
  
  himem_search();
  handle_chipset();

  for (i = 0; i < MAXMETEORS; ++i)
  {
    debug_msg("meteor%d: trying to find device\n", i);
    memset (&meteors[i], 0, sizeof (meteors[0]));
    if (!config_meteor (&meteors[i],i))
    {
      break;
    }
    meteors[i].video_dev = video_device_alloc();
    if ( meteors[i].video_dev == NULL )
    {
      err_msg ("meteor%d: Couldn't allocate video device.\n", i);
      unconfig_meteor ( &meteors[i] );
      return 0;
    }
    // on return from video_device_alloc the video_device is memset to 0
    // set up the video_device fields
    // TODO These could go in a meteor_init_video_device_fields() function
    strcpy(meteors[i].video_dev->name, "meteor_v4l2");
    meteors[i].video_dev->type2   = VFL_TYPE_GRABBER;
    meteors[i].video_dev->release = meteor_v4l2_release;
    meteors[i].video_dev->fops    = & meteor_v4l2_fops;
    video_set_drvdata( meteors[i].video_dev, &meteors[i] );

    if ( video_register_device ( meteors[i].video_dev, VFL_TYPE_GRABBER, -1) != 0 )
    {
      err_msg ("meteor%d: Couldn't register the driver.\n", i);
      video_device_release( meteors[i].video_dev );
      unconfig_meteor ( &meteors[i] );
      // this should not be needed ...
      // meteors[i].video_dev = NULL;
      return 0;
    }
    meteor_init_done( &meteors[i] );
    // meteors[i].is_registered = 1;
    debug_msg ("meteor%d: registered the driver.\n", i);
  }
  if (i == 0)
  {
    err_msg ("No meteor found.\n");
    return -ENODEV;		/* cleanup will not be called */
  }

  #ifdef METEOR_USE_WORKQUEUE
    meteor_wq_thread = create_workqueue("meteor_wq");
    if ( meteor_wq_thread == NULL ) {
	    err_msg("Work queue create error \n");
	    return -ENODEV; 
    }
  #endif

  masterclock = NULL; /* use the system clock as masterclock */

  return 0;
}


static void __exit
meteor_v4l2_exit (void)
{
  int i;

  for (i = 0; i < MAXMETEORS; ++i)
    if ( meteors[i].video_dev ) 
    {
      debug_msg("unregister meteor device %d\n", i);
      video_unregister_device( meteors[i].video_dev );
      unconfig_meteor ( &meteors[i] );
    }
  #ifdef METEOR_USE_WORKQUEUE
    destroy_workqueue( meteor_wq_thread );
  #endif
}

module_init( meteor_v4l2_init );
module_exit( meteor_v4l2_exit );

