/*
 *	meteor.h
 *
 *	Header file for the Matrox Meteor drivers.
 *
 *	This program is copyright 1999-2000
 *	by Olivier Carmona
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 */
#ifndef _METEOR_H
#define _METEOR_H

/* ---------------------------------------------------------------------------- */
/* Begin V4L2 extra stuff                                                       */
/* ---------------------------------------------------------------------------- */

#ifdef __KERNEL__

#include <linux/types.h>
#include <linux/version.h>

#include <linux/spinlock.h> // rwlock_t

#include <linux/poll.h>
#if (LINUX_VERSION_CODE >= 0x020300) || defined(CONFIG_FS_DEVFS)
#include <linux/devfs_fs_kernel.h>
#endif


#define V4L2_BUF_REQ_CONTIG      0x10000000
#define V4L2_BUF_TYPE_field      0x00001FFF  /* Type field mask */

#define V4L2_NUM_DEVICES    256

#else /* ! __KERNEL */

// compatibility
#define V4L2_BUF_TYPE_CAPTURE  V4L2_BUF_TYPE_VIDEO_CAPTURE

#endif /* __KERNEL__ */

// #define V4L2_FMT_FLAG_ODDFIELD         V4L2_FIELD_TOP
// #define V4L2_FMT_FLAG_EVENFIELD        V4L2_FIELD_BOTTOM
// #define V4L2_FMT_FLAG_NOT_INTERLACED   V4L2_FIELD_ALTERNATE
// #define V4L2_FMT_FLAG_INTERLACED       V4L2_FIELD_INTERLACED


struct v4l2_enumstd
{
	int         index;
	struct v4l2_standard    std;
	__u32           inputs;  /* set of inputs that support this standard */
	__u32           outputs; /* set of outputs that support this standard */
	__u32           reserved[2];
};

#ifdef CONFIG_V4L2_PERFORMANCE
struct v4l2_performance
{
	int frames;
	int framesdropped;
	__u64   bytesin;
	__u64   bytesout;
	__u32   reserved[4];
};

#define VIDIOC_G_PERF            _IOR  ('V', 20, struct v4l2_performance)
#endif // CONFIG_V4L2_PERFORMANCE



#ifdef __KERNEL__

typedef __s64 stamp_t;

extern struct page * v4l2_malloc_to_page( void * virt );

#ifdef METEOR_USE_LIST
  #include <linux/list.h>
#else
  /* Simple queue management */
  struct v4l2_q_node {
	struct v4l2_q_node * forw;
	struct v4l2_q_node * back;
  };

  struct v4l2_queue {
	struct v4l2_q_node * forw;
	struct v4l2_q_node * back;
	rwlock_t             qlock;
  };
#endif

struct v4l2_clock
{
    void    (*gettime)(stamp_t *);
};

/* ---------------------------------------------------------------------------- */
/* End V4L2 extra stuff                                                         */
/* ---------------------------------------------------------------------------- */

#define MAX_NUM_FRAMES	30

/* Miscelleneaous things */

// WAIT_JIFFS was 3 but with 2.6 HZ is 1000
#define WAIT_JIFFS 28   /* Should be about 28 milliseconds */
#define WAIT_JIFF1  9   /* one third of WAIT_JIFFS */

/* Jim Lowe suggests 32, or 16 if bus master slots > 4 */
/* TODO: test with different values */
#define METEOR_LATENCY 16


/* Meteor Flags */
#define	METEOR_FULL	        0x00000001
#define	METEOR_HAVE_EVEN	0x00100000
#define	METEOR_HAVE_ODD		0x00200000
#define	METEOR_HAVE_MASK	0x00300000
#define METEOR_RGB		0x20000000	/* meteor rgb unit */

/*
 *	Global device data structure
 */


struct meteor;/* forward reference */

struct video_decoder
{
	int	type;
	int	version;
	int	is_initialized;
	int	i2c_addr1;
	int	i2c_addr2;
	int	ntsc_hskip;
	int	ntsc_vskip;
	int	ntsc_width;
	int	ntsc_height;
	int	ntsc_field_order;
	int	pal_hskip;
	int	pal_vskip;
	int	pal_width;
	int	pal_height;
	int	pal_field_order;
	int	preferred_field;
	int	invert_vsync;
	int	hadjust;
	int	vadjust;

	int	num_inputs;
	int	input;
	// __u32	standards;
	// __u32	standard;
	v4l2_std_id	standards;
	v4l2_std_id	standard;
	// __u32	frame_period;
	struct v4l2_fract      frame_period;
	int	decoder_is_stable;
	__u32	decoder_stable_time;

	/*  Changable method functions helps us support multiple	*/
	/*  different types of video decoders easily			*/
	int	(*initialize)(struct meteor *dev);
	int	(*set_input)(struct meteor *dev, int x);
	int	(*set_oformat)(struct meteor *dev);
	int	(*set_standard)(struct meteor *dev, int x);
	int	(*set_vcrmode)(struct meteor *dev, int x);
	int	(*is_stable)(struct meteor *dev);
	int	(*init_reg)(struct meteor *dev);
	int	(*check)(struct meteor *dev);
};

#define MAXCONTROLS	4

struct video_source
{
	struct v4l2_input	input;
	int			control[MAXCONTROLS];
	int			vcrmode;
};

#define MAX_LOCKED_MEMORY	2000000

/* Maximum Number of Boards */
#define MTR_NBOARDS             4

/*  Number of video_source's */
#define METEOR_INPUT_DEV0	0	/* camera input 0 -- default */
#define METEOR_INPUT_DEV_RCA	METEOR_INPUT_DEV0
#define METEOR_INPUT_DEV1	1	/* camera input 1 */
#define METEOR_INPUT_DEV2	2	/* camera input 2 */
#define METEOR_INPUT_DEV3	3	/* camera input 3 */
#define METEOR_INPUT_DEV_SVIDEO	4       /* S-video input port */
#define METEOR_INPUT_DEV_RGB	5	/* for rgb version of meteor */
#define VSOURCE_COUNT		6

/*  Bus-master High memory frame buffer */
struct himem_node
{
	__u32	           addr;
	__u32	           len;
	int                used;
	struct himem_node  *next;
	struct himem_node  *previous;
};

/*  Image translations  */
struct translation
{
	int		type;
	int		width;
	int		height;
	__u8		*in;
	int		in_stride;
	__u8		*out;
	int		out_stride;
	int		output_size;
	int		output_is_user;
};
#define XLAT_NULL		0
#define XLAT_BGR32_TO_BGR24	8


/*  Per-open data for handling multiple opens on one device */
struct device_open
{
	int			isopen;
	int			noncapturing;
	struct meteor		*dev;
};
#define MAX_OPENS	3

/*  Streaming data buffer  */
struct stream_buffer
{
        #ifdef METEOR_USE_LIST
	  struct list_head      qnode;
	#else
	  struct v4l2_q_node	qnode;
	#endif
	struct v4l2_buffer	vidbuf;
	int			requested;
	__u8			*vaddress;  /* vmalloc() */
	struct himem_node	*himem_node;  /* get_free_page() */
};

/*
 *	Meteor device structure
 *
 *	One for each Meteor in the system.
 *	This structure holds all the global information the drivers
 *	need about each board.
 */
struct meteor
{
	// struct v4l2_device	v;
	struct video_device *   video_dev;
	char			shortname[16];

/*      Per-bus index number for each Meteors   */
        short                   index;
/*	Pointer to the pci_dev structure for this board	*/
	struct pci_dev		*pci;
	unsigned char		irq;
/*	Spinlock for this device to guarantee accesses are atomic	*/
	spinlock_t		lock;

/*
 *	V I D E O   C A P T U R E
 */
	// int			is_registered;  // (mc) not used
	int			open_count;
	struct device_open	open_data[MAX_OPENS];
	int			capturing_opens;

/*	Video Interrupts	*/
	int			ints_enabled;
	// (mc) these two are not used
	//      should use workqueue instead ?
	// struct tq_struct	tqnode_dpc;/* for BH */
	// struct timer_list	tlnode;/* for polling interrupts */
	wait_queue_head_t	new_video_frame;
	void			(*video_handler)(void *);

/*	Video decoder stuff		*/
	struct video_decoder	videc;
	struct video_source	source[VSOURCE_COUNT];
	int			source_width;
	int			source_height;
        __u32                   frame_period; /* in 100ns */
	char                    standard_name[24];

/*	Client capture format and capture modes	*/
	struct v4l2_pix_format	clientfmt;
	struct v4l2_captureparm	capture;
	int			input;
	__u32                   timeperframe; /* in 100ns */

/*	Hardware capture format */
	int			capture_bypp;
	int			capture_size;

/*	Capture buffer  */
	__u8			*capture_buffer;/* vmalloc() */
	int			capture_buffer_size;
	struct himem_node	*capture_himem_node;/* get_free_page() */


/*	Capture state	*/
	int			ready_to_capture;
	int			grabber_enabled;
	int			capture_completed;
	unsigned long		time_acquired;/* millisecond time stamp */
	int			streaming;
	struct stream_buffer	stream_buf[MAX_NUM_FRAMES];
	int			stream_buffers_requested;
	int			stream_buffers_mapped;
	int			stream_contig_map;
        #ifdef METEOR_USE_LIST
	  struct list_head      stream_q_capture;
	  struct list_head      stream_q_done;
        #else
	  struct v4l2_queue	stream_q_capture;
	  struct v4l2_queue	stream_q_done;
	#endif
	s64			stream_begin;
	unsigned long		stream_last_frame;
	struct stream_buffer	*stream_current_buffer;

/*	Image format conversions	*/
	struct translation	translation;
	__u8			*xlat_temp;/* vmalloc() */

#ifdef CONFIG_V4L2_PERFORMANCE
/*	Performance statistics	*/
	struct v4l2_performance perf;  // (mc) not used
#endif // CONFIG_V4L2_PERFORMANCE

/*      Frame counter for test images */
    	int h, m, s, f;

/*      Pointer to SAA 7196 registers */
	struct saa7116		*s7116;

/*      Maximum range-checkable size */
        unsigned long		dma_range;
/*      Number of fifo capture errors since open */
   	unsigned long		fifo_errors;
/* 	Number of DMA capture errors since open */
        unsigned long		dma_errors;
/* 	DMA address offset for even field */
   	unsigned long           dma_add_e[3];
/* 	DMA address offset for odd field */
   	unsigned long       	dma_add_o[3];
/* 	DMA stride for even field */
   	unsigned long       	dma_str_e[3];
/* 	DMA stride for odd field */
   	unsigned long       	dma_str_o[3];
/*      SAA 7196 register values */
   	unsigned char		saa7196_i2c[NUM_SAA7196_I2C_REGS];
/*      PCF8574 register values */
   	unsigned char		pcf_i2c[NUM_PCF8574_I2C_REGS];
/* 	BT254 register values */
   	unsigned char		bt254_reg[NUM_BT254_REGS];
/* 	TODO: completely suppress those historical flags */
   	unsigned long		flags;
};

/* Head of Himem List */
struct himem_node *himem_head;

#endif /* __KERNEL__ */

#endif /*_METEOR_H */
