/*
 *	saa71x6.h  -- SAA7116 (Video PCI Bridge) & SAA7196 (Video Decoder and Scaler) definitions
 *
 *	Copyright (C) 1999-2000  Olivier  Carmona (carmona@k-team.com)
 *	Copyright (C) 1996  Jim Bray (http://as220.org/jb).
 *
 *      Thanks to Hacene Machane from Matrox Imaging for 7116 specs.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _SAA_71X6_H
#define _SAA_71X6_H

/*
 * Definitions for the Philips SAA7196 digital video decoder,
 * scalar, and clock generator circuit (DESCpro).
 */
#define NUM_SAA7196_I2C_REGS	49
#define	SAA7196_I2C_ADDR_W	0x40
#define	SAA7196_I2C_ADDR_R	0x41

#define SAA7196_WRITE(mtr, reg, data) \
 saa7116_i2c(mtr, SAA7196_I2C_ADDR_W, reg, data), \
  mtr->saa7196_i2c[reg] = data; /* Update the saved image */
#define	SAA7196_READ(mtr) \
 saa7116_i2c(mtr, SAA7196_I2C_ADDR_R, 0x0, 0x0)

#define SAA7196_REG(mtr, reg) mtr->saa7196_i2c[reg]


#define SAA7196_IDEL	0x00	/* Increment delay */
#define SAA7196_HSB5	0x01	/* H-sync begin; 50 hz */
#define SAA7196_HSS5	0x02	/* H-sync stop; 50 hz */
#define SAA7196_HCB5	0x03	/* H-clamp begin; 50 hz */
#define SAA7196_HCS5	0x04	/* H-clamp stop; 50 hz */
#define SAA7196_HSP5	0x05	/* H-sync after PHI1; 50 hz */
#define SAA7196_LUMC	0x06	/* Luminance control */
#define SAA7196_HUEC	0x07	/* Hue control */
#define SAA7196_CKTQ	0x08	/* Colour Killer Threshold QAM (PAL, NTSC) */
#define SAA7196_CKTS	0x09	/* Colour Killer Threshold SECAM */
#define SAA7196_PALS	0x0a	/* PAL switch sensitivity */
#define SAA7196_SECAMS	0x0b	/* SECAM switch sensitivity */
#define SAA7196_CGAINC	0x0c	/* Chroma gain control */
#define SAA7196_STDC	0x0d	/* Standard/Mode control */
#define SAA7196_IOCC	0x0e	/* I/O and Clock Control */
#define SAA7196_CTRL1	0x0f	/* Control #1 */
#define SAA7196_CTRL2	0x10	/* Control #2 */
#define SAA7196_CGAINR	0x11	/* Chroma Gain Reference */
#define SAA7196_CSAT	0x12	/* Chroma Saturation */
#define SAA7196_CONT	0x13	/* Luminance Contrast */
#define SAA7196_HSB6	0x14	/* H-sync begin; 60 hz */
#define SAA7196_HSS6	0x15	/* H-sync stop; 60 hz */
#define SAA7196_HCB6	0x16	/* H-clamp begin; 60 hz */
#define SAA7196_HCS6	0x17	/* H-clamp stop; 60 hz */
#define SAA7196_HSP6	0x18	/* H-sync after PHI1; 60 hz */
#define SAA7196_BRIG	0x19	/* Luminance Brightness */
#define SAA7196_FMTS	0x20	/* Formats and sequence */
#define SAA7196_OUTPIX	0x21	/* Output data pixel/line */
#define SAA7196_INPIX	0x22	/* Input data pixel/line */
#define SAA7196_HWS	0x23	/* Horiz. window start */
#define SAA7196_HFILT	0x24	/* Horiz. filter */
#define SAA7196_OUTLINE	0x25	/* Output data lines/field */
#define SAA7196_INLINE	0x26	/* Input data lines/field */
#define SAA7196_VWS	0x27	/* Vertical window start */
#define SAA7196_VYP	0x28	/* AFS/vertical Y processing */
#define SAA7196_VBS	0x29	/* Vertical Bypass start */
#define SAA7196_VBCNT	0x2a	/* Vertical Bypass count */
#define SAA7196_VBP	0x2b	/* veritcal Bypass Polarity */
#define SAA7196_VLOW	0x2c	/* Colour-keying lower V limit */
#define SAA7196_VHIGH	0x2d	/* Colour-keying upper V limit */
#define SAA7196_ULOW	0x2e	/* Colour-keying lower U limit */
#define SAA7196_UHIGH	0x2f	/* Colour-keying upper U limit */
#define SAA7196_DPATH	0x30	/* Data path setting  */


/* Default saa7196 internal registers */
static u_char saa7196_i2c_default[NUM_SAA7196_I2C_REGS] =
{
/* SAA7196 I2C bus control                      */
/* BITS Function                                */
/* 00 */ 0x50,
	/* 	7:0  Increment Delay                         */
/* 01 */ 0x30,
	/* 	7:0  Horizontal Sync Begin for 50hz          */
/* 02 */ 0x00,
	/* 	7:0  Horizontal Sync Stop for 50hz           */
/* 03 */ 0xe8,
	/* 	7:0  Horizontal Sync Clamp Start for 50hz    */
/* 04 */ 0xb6,
	/* 	7:0  Horizontal Sync Clamp Stop for 50hz     */
/* 05 */ 0xf4,
	/* 	7:0  Horizontal Sync Start after PH1 for 50hz */
/* 06 */ 0x46,
	/*   	7    Input mode =0 CVBS, =1 S-Video 
		6    Pre filter
		5:4  Aperture Bandpass characteristics
		3:2  Coring range for high freq
		1:0  Aperture bandpass filter weights        */
/* 07 */ 0x80,
	/* 7:0  Hue                                     */
/* 08 */ 0x7f,
	/* 7:3  Colour-killer threshold QAM (PAL, NTSC) */
/* 09 */ 0x7f,
	/* 7:3  Colour-killer threshold SECAM           */
/* 0a */ 0x7f,
	/* 7:0  PAL switch sensitivity                  */
/* 0b */ 0x7f,
	/* 7:0  SECAM switch sensitivity                */
/* 0c */ 0x40,
	/*   	7  Colour-on bit
		6:5  AGC filter (0 slow/1 medium/2 fast)*/
/* 0d */ 0x84, // 84
	/*   	7  VTR/TV mode bit = 	1->VTR mode
		3    Realtime output mode select bit
		2    HREF position select
		1    Status byte select
		0    SECAM mode bit                          */
/* 0e */ 0x38,
	/*   	7  Horizontal clock PLL
 		5    Select interal/external clock source
		4    Output enable of Horizontal/Vertical sync
		3    Data output YUV enable
		2    S-VHS bit
		1    GPSW2
		0    GPSW1                                   */
/* 0f */ 0x50,
	/*   	7  Automatic Field detection
		6    Field Select 0 = 50hz, 1=60hz
		5    SECAM cross-colour reduction
		4    Enable sync and clamping pulse
		3:1  Luminance delay compensation            */
/* 10 */ 0x00,
	/*   	2  Select HREF Position
		1:0  Vertical noise reduction                */
/* 11 */ 0x2c,
	/* 	7:0  Chrominance gain control for QAM        */
/* 12 */ 0x40,
	/* 	7:0  Chrominance saturation control for VRAM port */
/* 13 */ 0x40,
	/* 	7:0  Luminance contract control for VRAM port */
/* 14 */ 0x34,
	/* 	7:0  Horizontal sync begin for 60hz          */
#if 0
/* 15 */ 0x0c,
	/* 	7:0  Horizontal sync stop for 60hz           */
/* 16 */ 0xfb,
	/* 	7:0  Horizontal clamp begin for 60hz         */
/* 17 */ 0xd4,
	/* 	7:0  Horizontal clamp stop for 60hz          */
/* 18 */ 0xec,
	/* 	7:0  Horizontal sync start after PH1 for 60hz */
#else
/* 15 */ 0x0a,
	/* 	7:0  Horizontal sync stop for ?           */
/* 16 */ 0xf4,
	/* 	7:0  Horizontal clamp begin for ?         */
/* 17 */ 0xce,
	/* 	7:0  Horizontal clamp stop for ?          */
/* 18 */ 0xf4,
	/* 	7:0  Horizontal sync start after PH1 for ? */
#endif
/* 19 */ 0x80,
	/* 	7:0  Luminance brightness control for VRAM port */
/* 1a */ 0x00,
/* 1b */ 0x00,
/* 1c */ 0x00,
/* 1d */ 0x00,
/* 1e */ 0x00,
/* 1f */ 0x00,
/* 20 */ 0x90, // 90
	/*	7  ROM table bypass switch
		6:5  Set output field mode
		4    VRAM port outputs enable
		3:2  First pixel position in VRO data
		1:0  FIFO output register select             */
/* 21 */ 0x80,
	/* 	7:0  [7:0] Pixel number per line on output   */
/* 22 */ 0x80,
	/* 	7:0  [7:0] Pixel number per line on input    */
/* 23 */ 0x03,
	/* 	7:0  [7:0] Horizontal start position of scaling win */
/* 24 */ 0x8a,
	/* 	7:5  Horizontal decimation filter
		4  [8] Horizontal start position of scaling win
		3:2  [9:8] Pixel number per line on input
		1:0  [9:8] Pixel number per line on output   */
/* 25 */ 0xf0,
	/* 	7:0  [7:0] Line number per output field      */
/* 26 */ 0xf0,
	/* 	7:0  [7:0] Line number per input field       */
/* 27 */ 0x0f,
	/* 	7:0  [7:0] Vertical start of scaling window  */
/* 28 */ 0x80, // 80: Adaptive filter 40: not so bad but 60 and 20 awfull
	/*   	7  Adaptive filter switch
		6:5  Vertical luminance data processing
		4    [8] Vertical start of scaling window 
		3:2  [9:8] Line number per input field
		1:0  [9:8] Line number per output field      */
/* 29 */ 0x16,
	/* 	7:0  [7:0] Vertical bypass start             */
/* 2a */ 0x00,
	/* 	7:0  [7:0] Vertical bypass count             */
/* 2b */ 0x00,
	/* 	4  [8] Vertical bypass start
	   	2  [8] Vertical bypass count
		0    Polarity, internally detected odd even flag */
/* 2c */ 0x80,
	/* 	7:0  Set lower limit V for colour-keying     */
/* 2d */ 0x7f,
	/* 	7:0  Set upper limit V for colour-keying     */
/* 2e */ 0x80,
	/* 	7:0  Set lower limit U for colour-keying     */
/* 2f */ 0x7f,
	/* 	7:0  Set upper limit U for colour-keying     */
/* 30 */ 0xcf //bf 
	/* 	7  VRAM bus output format
	   	6  Adaptive geometrical filter
	   	5  Luminance limiting value
		4  Monochrome and two's complement output data sel
		3  Line qualifier flag
		2  Pixel qualifier flag
		1  Transparent data transfer
		0  Extended formats enable bit               */
};


/*
 * Definitions for the Philips SAA7116 Digital Video to PCI 
 * Interface chip.
 */


/*
 * Bitfield words are each defined as a separate struct. This unclutters
 * the actual register struct definitions, and also enables use of gcc
 * constructor expressions for assignment, which is readable and results
 * in better object code than individual assignments to bitfields.
 *
 * Note bitfields build from LSB up.
 *
 * Slight variances from actual field width are made to use byte boundaries
 * when possible. Actual widths are noted in comments.
 */

/* Bitfields of the 7196 */

struct rt_mode {
  unsigned mode:8;		/* 0:7  Pixel format for even fields */
  unsigned route:24;		/* 8:31 Pixel router mode for even fields */
};

struct ff_trig {
  unsigned packed:7;		/* 0:7(0:6) Fifo trigger Packed Mode */
  unsigned :9;			/* 8:15(7:15) Unused */
  unsigned planar:7;		/* 16:31(16:22) Fifo trigger Planar Mode */
  unsigned :9;			/* 23:31 Unused */
};

struct i2c_stat {
  unsigned :9;			/* 0:8 Not needed */
  unsigned d_abort:1;		/* 9:9 Status of i2c write: 0==success */
  unsigned :14;			/* 10:23 Not needed */
  unsigned r_data:8;		/* 24:31 Data returned by i2c read */
};

struct i2c_cmd {
  unsigned w_data:8;		/* 0:7 Data to be written out */
  unsigned reg_addr:8;		/* 8:15 Register address in device space */
  unsigned dev_addr:8;		/* 16:23 16: 0==write; 17:23: Device address */
  unsigned xfer:1;		/* 24: Set to start write. Clears when done */
  unsigned :7;			/* 25:31 Unused */
};

#define v_struct volatile struct

/*
   RR Read/Reset: This register may be read, or be reset to 0 by
      writing a 1 into corresponding bit location.
   RS Read/Set  : This register may be read, or be set to 1 by
      writing a 1 into corresponding bit location.
*/
				/* (pointer-addition-values)hex-offset(s) */
struct saa7116 {
  volatile u_long dma_e[3];		/* (0-2)00-08:  31:2 Base address, even fields */
  volatile u_long dma_o[3];		/* (3-5)0c-14:  31:2 Base address, odd fields */
  volatile u_long stride_e[3];		/* (6-8)18-20:  15:2 Address stride, even fields */
  volatile u_long stride_o[3];		/* (9-11)24-2c: 15:2 Address stride, odd fields */
  v_struct rt_mode rt_mode_e;	/* (12)30:      31:0 Route and Mode for even fields */
  v_struct rt_mode rt_mode_o;	/* (13)34:      31:0 Route and Mode for odd fields */
  v_struct ff_trig fifo_trigger;	/* (14)38       22:0 Fifo triggers: Planar and Packed */
  volatile u_long field_toggle;		/* (15)3c:         2 Field togle bit (emulate interlaced with non 
                                                     interlaced video). */
  volatile u_long capt_ctl;	/* (16)40:        15 Enable range checking */
                                /*                14 Ignore FIFO Overflow (test only) */
				/*         RR     11 Odd field address error detected */
				/*         RR     10 Even field address error detected */
				/*         RR      9 Odd field FIFO overflow detected */
				/*         RR      8 Even field FIFO overflow detected */
				/*                 7 FIFO Enable */
				/*                 6 Reset the external video decoder */
				/*         RR      5 Odd field Done */
				/*         RR      4 Even field Done */
				/*         RS      3 Trigger a single odd field Capture */
				/*         RS      2 Trigger a single even field Capture */
				/*                 1 Trigger a continuous odd field capture */
				/*                 0 Trigger a continuous even field capture */
  u_long retry_wait;		/* (17)44        7:0 Number of clocks to wait after a slave 
                                                     disconnect before retrying */
  volatile u_long ints_ctl;	/* (18)48: RR     10 Interrupt mask start of field */
                                /*         RR      9 Interrupt mask end of odd field */
				/*         RR      8 Interrupt mask end of even field */
				/*         RR      2 Interrupt status start of field */
				/*         RR      1 Interrupt status end of odd field */
				/*         RR      0 Interrupt status end of even field */
  volatile u_long field_mask_e;		/* (19)4c */
  volatile u_long field_mask_o;		/* (20)50 */
  volatile u_long mask_lengths;		/* (21)54 */
  volatile u_long fifo_limits;		/* (22)58      22:16 FIFO Almost Empty Pointer: number of dwords needed
                                                     to trigger an end of PCI burst transfers. A good
						     value is 0x5. */
                                /*               6:0 FIFO ALmost Full Pointer number of dwords needed
                                                     to trigger a start of PCI burst transfers. A good
						     value is 0x7C.*/
  u_long i2c_clocks;		/* (23)5c      31:24 I2C Phase 4 clock cycles duration. */
                                /*             23:16 I2C Phase 3 clock cycles duration. */
                                /*              15:8 I2C Phase 2 clock cycles duration. */
                                /*               7:0 I2C Phase 1 clock cycles duration. */
  v_struct i2c_stat i2c_stat;	/* (24)60 i2c mode control and status */
  v_struct i2c_cmd i2c_cmd;	/* (25)64 Direct mode command and data */
  volatile u_long i2c_auto_0_e;		/* (26)68 */
  volatile u_long i2c_auto_2_e;		/* (27)6c */
  volatile u_long i2c_auto_4_e;		/* (28)70 */
  volatile u_long i2c_auto_6_e;		/* (29)74 */
  volatile u_long i2c_auto_0_o;		/* (30)78 */
  volatile u_long i2c_auto_2_o;		/* (31)7c */
  volatile u_long i2c_auto_4_o;		/* (32)80 */
  volatile u_long i2c_auto_6_o;		/* (33)84 */
  volatile u_long i2c_regs_enable;	/* (34)88 */
  volatile u_long dma_end_e;		/* (35)8c        23:2 Upper bound of the address window where the master 
                                                      may write to during even field DMA. Range enable
						      bit must be set. */
  volatile u_long dma_end_o;		/* (36)90        23:2 Upper bound of the address window where the master 
                                                      may write to during odd field DMA. Range enable
						      bit must be set. */
};
				


#undef v_struct

#endif

